"""Buy starting materials process."""

from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["BuyStartingMaterials"]


class BuyStartingMaterials(Process):
    """Class representing the purchase of a starting material from a supplier."""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(name=__name__)

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Supplier", bounds=CategoricalBounds(categories=[""])
        ),
        default_value=NominalCategorical(""),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)

__all__ = ["ArcMelting"]


class ArcMelting(Process):
    """Class representing arc melting"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Arc Melting",
        description="""One of the fabrication methods, arc melting""",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Argon Pressure",
            bounds=CategoricalBounds(["850-900", "900-950"]),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Diffusion",
            bounds=CategoricalBounds(categories=["Before Each Melt"]),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Ingot Location",
            bounds=RealBounds(0, 15, ""),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Initial Purging Times",
            bounds=RealBounds(0, 10, "hour"),
        ),
        default_value=NominalReal(0, "hour"),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Vacuum Before Melt",
            bounds=RealBounds(0, 15, ""),
        ),
        default_value=NominalReal(0, ""),
    )

    finalize_template(_ATTRS, TEMPLATE)



import json
import networkx as nx
from collections import defaultdict
import shutil
import os
import matplotlib.pyplot as plt
import argparse
import pathlib

from gemd.util.impl import recursive_foreach
from gemd.json import GEMDJson

from openmsimodel.utilities.argument_parsing import OpenMSIModelParser
from openmsimodel.utilities.runnable import Runnable
from openmsimodel.utilities.tools import read_gemd_data
from openmsimodel.graph.helpers import launch_graph_widget

import time


class OpenGraph(Runnable):
    """this class provides modules to build and visualize a networkx or graphviz object from gemd objects.
    By taking folder path containing GEMD thin JSON files, it establishes the relationships
    between them by interpreting their uuids/links, and flexibly produces anything from
    a svg image with simple labels, to a dot products containing all the gemd assets,
    from attributes, file links or tags.
    """

    ARGUMENT_PARSER_TYPE = OpenMSIModelParser

    IPYNB_FILENAME = pathlib.Path(
        pathlib.Path(__file__).parent.resolve() / "open_graph_config/template.ipynb"
    )
    CONFIG_FILENAME = pathlib.Path(
        pathlib.Path(__file__).parent.resolve() / "open_graph_config/.config"
    )

    # instance attributes
    # TODO: move build_graph function params to obj + store pygraphviz and networkx as obj attr
    def __init__(
        self, name, dirpath, output, layout, add_bidirectional_edges, restrictive=False
    ):
        self.name = name
        self.dirpath = pathlib.Path(dirpath) if not (type(dirpath) == list) else dirpath
        self.output = pathlib.Path(output)  # TODO: REQUIRES FULL PATH NOW; FIX
        self.restrictive = restrictive
        self.layout = layout
        self.add_bidirectional_edges = add_bidirectional_edges
        self.svg_path = None
        self.dot_path = None
        self.shapes = {"run": "circle", "spec": "rectangle", "template": "triangle"}

    # instance method
    def build_graph(
        self, assets_to_add, add_separate_node, which, update=True, uuid_to_track="auto"
    ):
        """creates a NetworkX graph representation of the GEMD relationships by reading every object
        generated by the GEMDEncoder object, storing all of its links by uid, and forming directed relationships,
        such as ingredient->process, or process->material
        It then allows filtering the objects mapped (i.e., removing spec or runs,
        measurements or ingredients) and saves a NetworkX graph in "dot" as .png

        Args:
            which (bool): to plot a graph of specs, runs or templates
            add_separate_node (bool): bool to determine whether or not to add assets as attribute of related node, or as separate node
            assets_to_add (dict): dict to determine which of attributes, tags and/or file links to add to model
            update (bool, optional): bool to determinate updating instance variable svg_path and dot_path. Defaults to True.
            uuid_to_track (str, optional): _description_. Defaults to "auto".

        Returns:
            NetworkX Graph G_nx: graph containing knowledge in question, with uuids as name
            PyGraphVIZ Graph G_pviz_relabeled: graph containing highlighted knowledge in question, with individual names assigned to uuids
            dict name_mapping: mapping from uuid to name
        """

        print(
            "-- Building {}s of {}".format(
                which,
                self.dirpath
                if not type(self.dirpath) == list
                else f"list with {len(self.dirpath)} items",
            )
        )
        G_nx = nx.DiGraph(name=self.name)
        object_mapping = defaultdict()
        name_mapping = defaultdict()
        encoder = GEMDJson()
        nb_disregarded = 0

        gemd_objects, gemd_paths = read_gemd_data(self.dirpath, encoder)
        # gemd_objects = gemd_objects[:2000]

        if len(gemd_objects) == 0:
            return

        # adding objects to graph one by one
        for i, obj_data in enumerate(gemd_objects):
            obj_type = obj_data["type"]
            obj_state = obj_type.split("_")[-1]
            if not (uuid_to_track in obj_data["uids"].keys()):
                continue
            obj_uid = obj_data["uids"][uuid_to_track]
            obj_name = obj_data["name"]
            name_mapping[obj_uid] = "{} [{}, {}]".format(obj_name, obj_uid, obj_type)
            self.handle_gemd_obj(
                G_nx,
                obj_uid,
                obj_data,
                obj_type,
                obj_state,
                which,
                assets_to_add,
                add_separate_node,
            )
            try:
                path = gemd_paths[i]
                self.add_to_graph(
                    G_nx, obj_uid, "source", path.name, add_separate_node=False
                )
            except IndexError:
                continue
            if i % 1000 == 0:
                print("{} gemd objects processed...".format(i))
        print("Done.")

        # relabelling according to uid -> name
        print(name_mapping)
        if name_mapping:
            relabeled_G_nx = nx.relabel_nodes(G_nx, name_mapping)

        # converting to grapviz
        # relabeled_G_gviz = self.map_to_graphviz(relabeled_G_nx)
        relabeled_G_gviz = None

        # # plotting
        dot_path, svg_path, graphml_path = self.save_graph(
            self.output,
            relabeled_G_nx,
            relabeled_G_gviz,
            name="{}_{}".format(self.name, which),
        )
        if update:
            self.update_paths(svg_path, dot_path)

        # info
        self.diagnostics(relabeled_G_nx, gemd_objects, nb_disregarded)

        return relabeled_G_nx, relabeled_G_gviz, name_mapping

    def handle_gemd_obj(
        self,
        G,
        uid,
        obj_data,
        obj_type,
        obj_state,
        which,
        assets_to_add,
        add_separate_node,
    ):
        """method to handle the addition of a gemd object

        Args:
            G (NetworkX graph): graph
            uid (str): uid of current object
            obj_data (dict): data of current object
            obj_type (str): type of current object
            which (bool): to plot a graph of specs, runs or templates, or all
            assets_to_add (dict): dict to determine which of attributes, tags and/or file links to add to model
            add_separate_node (bool):  bool to determine whether or not to add assets as attribute of related node, or as separate node
        """
        if obj_type.startswith("process"):
            if obj_type.endswith(which) or which == "all":
                G.add_node(uid, color="red", shape=self.shapes[obj_state])
                self.add_gemd_assets(
                    G,
                    uid,
                    obj_data,
                    obj_type,
                    assets_to_add,
                    add_separate_node,
                )
        elif obj_type.startswith("ingredient"):  # TODO if node doesn't exist, create?
            if obj_type.endswith(which) or which == "all":
                G.add_node(uid, color="blue", shape=self.shapes[obj_state])
                process = obj_data["process"]["id"]
                G.add_edge(uid, process)
                if self.add_bidirectional_edges:
                    G.add_edge(process, uid)
                if (
                    not self.restrictive
                ):  # no material -> ingredient = helps separate and restrict
                    if "material" in obj_data and obj_data["material"]:
                        material = obj_data["material"]["id"]
                        G.add_edge(material, uid)
                        if self.add_bidirectional_edges:
                            G.add_edge(uid, material)
                self.add_gemd_assets(
                    G,
                    uid,
                    obj_data,
                    obj_type,
                    assets_to_add,
                    add_separate_node,
                )
        elif obj_type.startswith("material"):
            if obj_type.endswith(which) or which == "all":
                G.add_node(uid, color="green", shape=self.shapes[obj_state])
                self.add_gemd_assets(
                    G,
                    uid,
                    obj_data,
                    obj_type,
                    assets_to_add,
                    add_separate_node,
                )
                # if not self.restrictive:
                if "process" in obj_data and obj_data["process"]:
                    process = obj_data["process"]["id"]
                    G.add_edge(process, uid)  # ?
                    if self.add_bidirectional_edges:
                        G.add_edge(uid, process)
        elif obj_type.startswith("measurement"):
            if obj_type.endswith(which) or which == "all":
                G.add_node(uid, color="purple", shape=self.shapes[obj_state])
                self.add_gemd_assets(
                    G,
                    uid,
                    obj_data,
                    obj_type,
                    assets_to_add,
                    add_separate_node,
                )
                # if not self.restrictive:
                if "material" in obj_data and obj_data["material"]:
                    material = obj_data["material"]["id"]
                    G.add_edge(uid, material)
                    if self.add_bidirectional_edges:
                        G.add_edge(material, uid)

        if which == "all":
            if (
                obj_type.startswith("condition")
                or obj_type.startswith("parameter")
                or obj_type.startswith("property")
            ):
                G.add_node(uid, color="black", shape="trapezium")
                self.add_gemd_assets(
                    G,
                    uid,
                    obj_data,
                    obj_type,
                    assets_to_add,
                    add_separate_node,
                )
                # TODO: add add gemd_assets
            # adding attribute templates from object templates
            elif obj_type.endswith("template"):
                if "parameters" in obj_data and obj_data["parameters"]:
                    for parameter in obj_data["parameters"]:
                        G.add_edge(uid, parameter[0]["id"])
                if "conditions" in obj_data and obj_data["conditions"]:
                    for condition in obj_data["conditions"]:
                        G.add_edge(uid, condition[0]["id"])
                if "properties" in obj_data and obj_data["properties"]:
                    for prop in obj_data["properties"]:
                        G.add_edge(uid, prop[0]["id"])

            if obj_type.endswith("run"):
                if "spec" in obj_data and "id" in obj_data["spec"]:
                    spec_id = obj_data["spec"]["id"]
                    G.add_edge(uid, spec_id)
                    if self.add_bidirectional_edges:
                        G.add_edge(spec_id, uid)
            elif obj_type.endswith("spec"):
                if (
                    "template" in obj_data
                    and obj_data["template"] is not None
                    and "id" in obj_data["template"]
                ):
                    template_id = obj_data["template"]["id"]
                    G.add_edge(uid, template_id)
                    if self.add_bidirectional_edges:
                        G.add_edge(template_id, uid)

    def add_gemd_assets(
        self,
        G,
        uid,
        obj_data,
        obj_type,
        assets_to_add,
        add_separate_node,
    ):
        self.add_to_graph(G, uid, "uuid", uid, add_separate_node=False)
        self.add_to_graph(G, uid, "type", obj_type, add_separate_node=False)
        if self.layout == "raw":
            self.add_to_graph(G, uid, "object", str(obj_data), add_separate_node=False)
        elif self.layout == "visualization":
            if assets_to_add["add_attributes"] and not (obj_type.endswith("template")):
                if "parameters" in obj_data:
                    self.handle_gemd_value(
                        G, uid, obj_data["parameters"], add_separate_node
                    )
                if "properties" in obj_data:
                    self.handle_gemd_value(
                        G, uid, obj_data["properties"], add_separate_node
                    )
                if "conditions" in obj_data:
                    self.handle_gemd_value(
                        G, uid, obj_data["conditions"], add_separate_node
                    )
            if assets_to_add["add_file_links"] and "file_links" in obj_data:
                self.handle_gemd_value(
                    G, uid, obj_data["file_links"], add_separate_node
                )
            if assets_to_add["add_tags"] and "tags" in obj_data:
                self.handle_gemd_value(G, uid, obj_data["tags"], add_separate_node)

    def handle_gemd_value(self, G, uid, assets, add_separate_node):
        # TODO: add pointing to templates?
        for att in assets:
            if type(att) in [list]:
                continue
            if type(att) in [str]:
                if "::" in att:  # is a gemd tag
                    self.add_to_graph(G, uid, "tags", att, add_separate_node=False)
            elif att["type"]:  # is a gemd object
                # reading gemd file links
                if att["type"] == "file_link":
                    self.add_to_graph(
                        G, uid, "file_links", att["url"], add_separate_node=False
                    )
                    continue
                # reading gemd attributes
                if att["type"] == "property_and_conditions":
                    value = att["property"]["value"]
                    att_name = att["property"]["name"]
                else:
                    value = att["value"]
                    att_name = att["name"]
                if value["type"] == "nominal_real":
                    node_name = "{}, {} {}".format(
                        att_name, value["nominal"], value["units"]
                    )
                elif value["type"] == "nominal_integer":
                    node_name = "{}, {}".format(att_name, value["nominal"])
                elif value["type"] == "uniform_real":
                    node_name = "{}, {}-{} {}".format(
                        att_name,
                        value["lower_bound"],
                        value["upper_bound"],
                        value["units"],
                    )
                elif (
                    value["type"] == "uniform_integer"
                ):  # FIXME same as above without units
                    node_name = "{}, {}-{}".format(
                        att_name,
                        value["lower_bound"],
                        value["upper_bound"],
                    )
                elif (
                    value["type"] == "empirical_formula"
                ):  # FIXME same as above without units
                    node_name = "{}, {}, {}".format(
                        att_name,
                        value["formula"],
                        value["type"],
                    )
                elif (
                    value["type"] == "normal_real"
                ):  # FIXME same as above without units
                    node_name = "{}, mean {}, std {}, {}, {}".format(
                        att_name,
                        value["mean"],
                        value["std"],
                        value["units"],
                        value["type"],
                    )
                elif value["type"] == "nominal_categorical":
                    node_name = "{}, {}".format(att_name, value["category"])
                elif value["type"] == "nominal_composition":
                    node_name = "{}, {}".format(att_name, value["quantities"])

                self.add_to_graph(G, uid, att_name, node_name, add_separate_node)

    def add_to_graph(self, G, uid, att_name, node_name, add_separate_node):
        if add_separate_node == True:  # add as a separate node
            G.add_node(node_name, s="rectangle", color="orange")
            G.add_edge(uid, node_name)
            if self.add_bidirectional_edges:
                G.add_edge(node_name, uid)
        else:  # add as an attribute of the node
            if uid in G.nodes.keys():
                if att_name in G.nodes[uid].keys():  # already exists, append to it
                    if type(G.nodes[uid][att_name]) == dict:
                        count = len(G.nodes[uid][att_name])
                        G.nodes[uid][att_name][count] = node_name
                    return
                if att_name in ["file_links", "tags"]:
                    G.nodes[uid][att_name] = {0: node_name}
                else:
                    G.nodes[uid][att_name] = node_name

    def diagnostics(self, G, gemd_objects, nb_disregarded):
        print("-- Analysis --")
        if not self.add_bidirectional_edges:
            print("cycles in the graph: {}".format(list(nx.simple_cycles(G))))
        print(
            "nb of disregarded elements: {}/{}".format(
                nb_disregarded, len(gemd_objects)
            )
        )
        subgraphs = [G.subgraph(c).copy() for c in nx.strongly_connected_components(G)]
        # print("number of strongly connected components: {}".format(len(subgraphs)))
        print("nb of isolates: {}".format(nx.number_of_isolates(G)))

    @classmethod
    def launch(cls, dot_path, from_command_line=False):
        if from_command_line:
            with open(cls.CONFIG_FILENAME, "w") as f:
                f.write(dot_path)
            os.system(
                "jupyter notebook --notebook-dir={}".format(cls.IPYNB_FILENAME.parent)
            )
            return None
        else:
            launch_graph_widget(dot_path)

    def update_paths(self, svg_path, dot_path):
        self.svg_path = svg_path
        self.dot_path = dot_path

    @classmethod
    def map_to_graphviz(cls, G):
        """helper method to map NetworkX graph to Graphviz graph

        Args:
            G (_type_): _description_
            name_mapping (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """

        G = nx.nx_agraph.to_agraph(G)
        # G.node_attr.update(nodesep=0.4)
        # G.node_attr.update(ranksep=1)
        G.layout(prog="dot")
        return G

    @classmethod
    def slice_subgraph(cls, G, uuid, funcs, add_current=True):
        """applies paseed function(s) to graph object of interest with uuid=uuid.
        If elements are found to match the criteria, a subgraph containing all those elements is returned

        Args:
            G (NetworkX graph): knowledge graph in questoin
            uuid (str): uuid of current element of interest on which the functions are applied
            funcs (list): list of function(s) to apply to graph
            add_current (bool, optional): whether or not to add the current element of interest. Defaults to True.

        G (NetworkX graph): Graph to save
        """
        els = set()
        for func in funcs:
            els = els.union(func(G, uuid))
        if add_current:
            els.add(uuid)
        return G.subgraph(els)

    @classmethod
    def return_uuid(cls, identifier):
        """return the identifier of interest.

        Args:
            identifier (str): identifier of object of interest

        Returns:
            str: identifier
        """
        return identifier

    @classmethod
    def extract_subgraph(cls, G, identifier, func):
        """extract subgraph from graph knowledge, based on functions applied to element of interest to filter in additional desired elements.
        Examples includes neighbords, descendants, ancestors, etc.

        Args:
            G (NetworkX graph): Graph to save
            identifier (str): uuid or identifier of element of interest
            func (func): function to determine whether graph element should be added to subgraph or not

        Returns:
            NetworkX graph: subgraph filtered based on passed criteria
        """
        uuid = cls.return_uuid(identifier)
        return cls.slice_subgraph(G, uuid, func)

    @classmethod
    def save_graph(cls, dest, G_nx, G_gviz, name):
        """class method to save Graphviz graph.

        Args:
            dest (Pathlib.Path): path where to save the graph
            G_nx (Networkx graph): Networkx version of graph
            G_gviz (Graphviz graph): Graphviz version of graph
            name (str): name of file to save graph to

        Returns:
            str: paths to, respectively, the dot and svg files
        """
        svg_path = os.path.join(dest, "{}.svg".format(name))
        dot_path = os.path.join(dest, "{}.dot".format(name))
        graphml_path = os.path.join(dest, "{}.graphml".format(name))

        # # writing svg file
        # print("Dumping svg...")
        # start = time.time()
        # G_gviz.draw(svg_path)
        # plt.close()
        # end = time.time()
        # print(f"Time elapsed: {end - start}")

        # # writing dot file
        # print("Dumping dot...")
        # start = time.time()
        # with open(dot_path, "w") as f:
        #     f.write(str(G_gviz))
        # end = time.time()
        # print(f"Time elapsed: {end - start}")

        # writing graphml
        print("Dumping graphml...")
        start = time.time()

        def dicts_to_str(G):
            for node_name in G.nodes:
                if "tags" in G.nodes[node_name] and G.nodes[node_name]["tags"]:
                    G.nodes[node_name]["tags"] = str(G.nodes[node_name]["tags"])
                if (
                    "file_links" in G.nodes[node_name]
                    and G.nodes[node_name]["file_links"]
                ):
                    G.nodes[node_name]["file_links"] = str(
                        G.nodes[node_name]["file_links"]
                    )
            return G

        nx.write_graphml_lxml(dicts_to_str(G_nx), graphml_path)
        end = time.time()
        print(f"Time elapsed: {end - start}")

        print(
            "-- Saved graph to {}, {} and {}".format(dot_path, svg_path, graphml_path)
        )
        return dot_path, svg_path, graphml_path

    @classmethod
    def get_argument_parser(cls, *args, **kwargs):
        parser = cls.ARGUMENT_PARSER_TYPE(*args, **kwargs)
        cl_args, cl_kwargs = cls.get_command_line_arguments()
        parser.add_arguments(*cl_args, **cl_kwargs)
        return parser

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "name",
            "dirpath",
            "which",
            "identifier",
            "launch_notebook",
            "add_attributes",
            "add_file_links",
            "add_tags",
            "add_separate_node",
            "add_bidirectional_edges",
            "layout",
            "a",
            "d",
            "uuid_to_track",
            "output",
        ]
        kwargs = {**superkwargs}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        """
        Run a :class:`~OpenGraph` directly from the command line
        Calls :func:`~reconstruct` on a :class:`~OpenGraph` defined by
        command line (or given) arguments
        :param args: the list of arguments to send to the parser instead of getting them from sys.argv
        :type args: list, optional
        """
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        viewer = cls(
            args.name,
            args.dirpath,
            args.output,
            args.layout,
            args.add_bidirectional_edges,
        )
        assets_to_add = {
            "add_attributes": args.add_attributes,
            "add_file_links": args.add_file_links,
            "add_tags": args.add_tags,
        }
        G, relabeled_G_gviz, name_mapping = viewer.build_graph(
            assets_to_add=assets_to_add,
            add_separate_node=args.add_separate_node,
            which=args.which,
            uuid_to_track=args.uuid_to_track,
            # layout=args.layout,
            # add_bidirectional_edges=args.add_bidirectional_edges
        )

        # reduces to elements with identifier and related nodes
        if args.identifier:
            functions = []
            if args.d:
                functions.append(nx.descendants)
            if args.a:
                functions.append(nx.ancestors)
            identifier_G = cls.extract_subgraph(G, args.identifier, func=functions)
            identifier_G = cls.map_to_graphviz(identifier_G, name_mapping)
            identifier_G_dot_path, _ = cls.save_graph(
                args.output, identifier_G, "{}".format(args.identifier)
            )

        # launches interactive notebook
        if args.launch_notebook:
            if args.identifier:
                viewer.launch(identifier_G_dot_path, from_command_line=True)
            else:
                viewer.launch(viewer.dot_path, from_command_line=True)


def main(args=None):
    """
    Main method to run from command line
    """
    OpenGraph.run_from_command_line(args)


if __name__ == "__main__":
    main()



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
    _validate_temp_keys,
)


class AggregateTravelerSamples(Process):
    """Class representing aggegation of traveler samples"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Aggregate Traveler Samples",
    )

    _ATTRS: ClassVar[AttrsDict] = _validate_temp_keys(TEMPLATE)

    finalize_template(_ATTRS, TEMPLATE)



"""Base class for processes."""

from typing import ClassVar, Optional

from gemd import ProcessTemplate, ProcessSpec, ProcessRun
from gemd.entity.util import make_instance

from .process_or_measurement import ProcessOrMeasurement

__all__ = ["Process"]


class Process(ProcessOrMeasurement):
    """
    Base element for processes.

    """

    _TempType = ProcessTemplate
    _SpecType = ProcessSpec
    _RunType = ProcessRun

    TEMPLATE: ClassVar[ProcessTemplate]

    @property
    def spec(self) -> ProcessSpec:
        """The underlying process spec."""
        return self._spec

    @property
    def run(self) -> ProcessRun:
        """The underlying process run."""
        return self._run

    @classmethod
    def from_spec_or_run(
        cls,
        name: str,
        *,
        notes: Optional[str] = None,
        spec: ProcessSpec = None,
        run: ProcessRun = None
    ) -> "Process":
        """
        Instantiate a `Process` from a spec or run with appropriate validation.

        Note that the spec's template will be set to the class template,
        and the run's spec will be set to this spec.
        """

        if run is not None:
            if spec is None:
                spec = run.spec
        else:  # already checked that both can't be none, so only one scenario left
            if spec is None:
                raise ValueError("At least one of spec or run must be given.")
        template = spec.template

        process = cls(name, notes=notes, template=template)

        if spec is not None:
            if not isinstance(spec, ProcessSpec):
                raise TypeError("spec must be a ProcessSpec.")

            process._spec = spec

            process.spec.name = name
            process.spec.notes = notes
            # process.spec.template = cls.TEMPLATE

            process.update_conditions(which="spec")
            process.update_parameters(which="spec")

        if run is not None:
            if not isinstance(run, ProcessRun):
                raise TypeError("run must be a ProcessRun.")

            process._run = run

            process.run.name = name
            process.run.notes = notes
            process.run.spec = process.spec

            process.update_conditions(which="run")
            process.update_parameters(which="run")

            source = process.get_source()
            if source:
                process.set_source(
                    email=source["performed_by"], iso_date=source["performed_date"]
                )

        else:
            process._run = make_instance(process.spec)

        return process



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["InferCompositions"]


class InferCompositions(Process):
    """Class representing the inference of compositions via BO"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Infer Compositions",
        description="""Applying Bayezian Optimization based on multi-inputs/multi-output
        objectives to generated new composition space, with data extracted 
        from summary sheet aggregated by various experimental teams
        """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    MeasurementTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)
from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Tensile"]


class Tensile(Measurement):
    """Class representing Tensile as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="Tensile", description="Tensile measurement"
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Elastic Modulus, GPa", bounds=RealBounds(0, 500, "GPa")
        ),
        # default_value=NominalReal(0, "GPa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Yield Strength, MPa", bounds=RealBounds(-10000, 10000, "MPa")
        ),
        # default_value=NominalReal(0, "MPa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(name="UTS, MPa", bounds=RealBounds(-10000, 10000, "MPa")),
        # default_value=NominalReal(0, "MPa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(name="Elongation, %", bounds=RealBounds(0, 100, "")),
        # default_value=NominalReal(0, "MPa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Strain Hardening, MPa", bounds=RealBounds(-10000, 10000, "MPa")
        ),
        # default_value=NominalReal(0, "MPa"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(name="dUTS_dY", bounds=RealBounds(0, 100, "")),
        # default_value=NominalReal(0, "MPa"),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
)

from openmsimodel.entity.base.process import Process
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["SelectComposition"]


class SelectComposition(Process):
    """Class representing the selection of a composition among the choices"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Select Composition",
        description="""Selecting a composition from suggested compositions by Bayesian inference
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    finalize_template(_ATTRS, TEMPLATE)



# imports
import pkg_resources, subprocess, unittest, sys, importlib
from pathlib import Path
import os

sys.path.insert(0, "..")

from data.subclassing.arcmelting import ArcMelting
from openmsimodel.entity.impl import assign_uuid

from openmsimodel.stores.gemd_template_store import GEMDTemplateStore, stores_config

from openmsimodel.entity.base.base_element import BaseElement
from openmsimodel.entity.base.material import Material
from openmsimodel.entity.base.process import Process
from openmsimodel.entity.base.ingredient import Ingredient
from openmsimodel.entity.base.measurement import Measurement


from gemd import (
    NominalCategorical,
    NominalReal,
    CategoricalBounds,
    RealBounds,
    ProcessRun,
    ProcessSpec,
    ProcessTemplate,
    MaterialTemplate,
    MeasurementTemplate,
    Parameter,
    ParameterTemplate,
    Property,
    PropertyTemplate,
    Condition,
    ConditionTemplate,
)
from gemd.json import GEMDJson
from gemd.entity.util import make_instance


class TestStores(unittest.TestCase):
    # TODO: test define attributes when the same attribute is repassed

    # def test_global_store_ids_list(self):
    # self.assertEqual(len(template_store_ids), 2)
    # self.assertTrue("global" in template_store_ids)
    # self.assertTrue("test" in template_store_ids)

    # testing type of initialization 1 (see 1b for type of initialization 2)

    dummy_root = Path(__file__).parent.parent / "data/stores/dummy_store"
    # dummy_template_store = GEMDTemplateStore(id="dummy", load_all_files=False)

    stores_config.designated_store_id = "dummy"
    stores_config.deploy_store("dummy")
    # stores_config.all_template_stores["test"].root = dummy_root
    # stores_config.all_template_stores["test"].register_all_templates_from_store()

    # self.assertEqual(len(template_store_ids), 3)
    # self.assertTrue("dummy" in template_store_ids)

    def test_stores(self):
        """testing stores"""

        #######
        # run_all_tests will run test_entity_base_instantiation first, which populates the stores already wit ArcMelting initiations for example,
        # in which case both declarations below should return a warning for template already in file but still have the same template by uid
        # Expected: should get warnings with existing

        # with self.assertWarns(ResourceWarning):
        a = ArcMelting("arc melting")
        first_auto_uid = ArcMelting.TEMPLATE.uids["auto"]
        a_2 = ArcMelting("arc melting")
        self.assertEquals(first_auto_uid, a_2.TEMPLATE.uids["auto"])
        # if (
        #     global_template_store.id == "global"
        # ):  # if ran on its own TODO: review bc global template reads from gloabl stores
        # if (
        #     len(stores_config.all_template_stores.keys()) == 1
        #     and "global" in stores_config.all_template_stores.keys()
        # ):
        #     self.assertEqual(a.TEMPLATE_WRAPPER["global"].from_file, False)
        #     self.assertEqual(a.TEMPLATE_WRAPPER["global"].from_memory, False)
        #     self.assertEqual(a.TEMPLATE_WRAPPER["global"].from_subclass, True)
        #     self.assertEqual(a.TEMPLATE_WRAPPER["global"].from_store, False)
        if (
            "test" in stores_config.all_template_stores.keys()
        ):  # when all tests are ran. in this case, other tests are ran prior to this one and affect the store
            self.assertEqual(a.TEMPLATE_WRAPPER["test"].from_file, False)
            self.assertEqual(a.TEMPLATE_WRAPPER["test"].from_memory, False)
            self.assertEqual(a.TEMPLATE_WRAPPER["test"].from_subclass, False)
            self.assertEqual(a.TEMPLATE_WRAPPER["test"].from_store, True)
            self.assertEqual(a_2.TEMPLATE_WRAPPER["test"].from_file, False)
            self.assertEqual(a_2.TEMPLATE_WRAPPER["test"].from_memory, False)
            self.assertEqual(a_2.TEMPLATE_WRAPPER["test"].from_subclass, False)
            self.assertEqual(a_2.TEMPLATE_WRAPPER["test"].from_store, True)
        del a
        del a_2
        # self.assertNotEquals(first_auto_uid, a.TEMPLATE.uids["auto"])

    def test_dummy_store(self):
        ###### testing property setter
        stores_config.all_template_stores["dummy"].root = self.dummy_root
        self.assertEquals(
            stores_config.all_template_stores["dummy"].root,
            Path(__file__).parent.parent / "data/stores/dummy_store",
        )

        ##### testing initializing store
        stores_config.all_template_stores["dummy"].initialize_store()

        for root, dirs, files in os.walk(self.dummy_root):
            self.assertTrue(len(files) == 1)
            self.assertEqual(
                os.path.join(root, files[0]),
                str(stores_config.all_template_stores["dummy"].registry_path),
            )
            for dir in dirs:
                full_dir = Path(os.path.join(root, dir))
                self.assertTrue(
                    full_dir
                    in stores_config.all_template_stores["dummy"].store_folders.values()
                )
            break

        ###### testing register_template on a process template
        template = ProcessTemplate("dummy process template 1")
        _type = type(template)
        process_templates_destination = stores_config.all_template_stores[
            "dummy"
        ].store_folders[_type]

        with self.assertRaises(TypeError):
            stores_config.all_template_stores["dummy"].register_new_template(
                "str", from_store=False
            )

        self.assertEqual(len(os.listdir(process_templates_destination)), 0)
        stores_config.all_template_stores["dummy"].register_new_template(
            template, from_store=False
        )
        self.assertEqual(len(os.listdir(process_templates_destination)), 1)
        self.assertEqual(
            len(stores_config.all_template_stores["dummy"]._object_templates[_type]), 1
        )
        self.assertEqual(
            stores_config.all_template_stores["dummy"]
            ._object_templates[_type]["dummy process template 1"]
            .template.name,
            "dummy process template 1",
        )
        self.assertTrue(
            "auto"
            in stores_config.all_template_stores["dummy"]
            ._object_templates[_type]["dummy process template 1"]
            .template.uids.keys()
        )

        with self.assertWarns(
            ResourceWarning
        ):  # make sure it hasn't been added, as the already registered template should have been used
            stores_config.all_template_stores["dummy"].register_new_template(
                template, from_store=False
            )
        self.assertEqual(
            len(stores_config.all_template_stores["dummy"]._object_templates[_type]), 1
        )

        ###### testing register_template on a process template + attribute template (each individually)
        pt = ParameterTemplate(
            "dummy parameter template 1", bounds=RealBounds(0, 48, "hr")
        )
        template = ProcessTemplate("dummy process template 2", parameters=[pt])
        _type = type(pt)
        parameter_templates_destination = stores_config.all_template_stores[
            "dummy"
        ].store_folders[_type]

        self.assertEqual(len(os.listdir(parameter_templates_destination)), 0)

        stores_config.all_template_stores["dummy"].register_new_template(
            template, from_store=False
        )  # template
        stores_config.all_template_stores["dummy"].register_new_template(
            pt, from_store=False
        )  # attr

        self.assertEqual(len(os.listdir(parameter_templates_destination)), 1)  # + 1
        self.assertEqual(
            len(stores_config.all_template_stores["dummy"]._attribute_templates[_type]),
            1,
        )

        self.assertEqual(len(os.listdir(process_templates_destination)), 2)  # + 1
        self.assertEqual(
            len(
                stores_config.all_template_stores["dummy"]._object_templates[
                    type(template)
                ]
            ),
            2,
        )

    def test_dummy_store_with_BaseElement(self):
        # test store is manipulated in test_entity_base_instantiation prior (expected and useful for testing across files)
        test_processes_count = len(
            stores_config.all_template_stores["test"]._object_templates[ProcessTemplate]
        )
        test_params_count = len(
            stores_config.all_template_stores["test"]._attribute_templates[
                ParameterTemplate
            ]
        )
        self.assertEqual(test_processes_count, 8)
        self.assertEqual(test_params_count, 5)

        # adding dummy store
        self.assertEqual(len(stores_config.all_template_stores), 1)
        stores_config.all_template_stores["dummy"] = stores_config.all_template_stores[
            "dummy"
        ]
        self.assertEqual(len(stores_config.all_template_stores), 2)

        ##### 1) a) instantiation of base node object, which should register an object template and attribute template
        p = Process(
            "dummy process 1",
            template=ProcessTemplate(
                "dummy process template 3",
                parameters=[
                    ParameterTemplate(
                        "dummy parameter template 2", bounds=RealBounds(0, 10, "m")
                    )
                ],
            ),
        )

        ######
        def helper_1(one, two, three, four):
            self.assertEqual(one, two)
            self.assertEqual(three, four)

        def reusable_tests():
            # test asserts
            test_processes = stores_config.all_template_stores[
                "test"
            ]._object_templates[ProcessTemplate]
            test_processes_destination = stores_config.all_template_stores[
                "test"
            ].store_folders[ProcessTemplate]
            helper_1(  # due to 1a), gets one more
                len(test_processes) - test_processes_count,
                1,
                len(os.listdir(test_processes_destination)) - test_processes_count,
                1,
            )
            test_parameters = stores_config.all_template_stores[
                "test"
            ]._attribute_templates[ParameterTemplate]
            test_parameters_destination = stores_config.all_template_stores[
                "test"
            ].store_folders[ParameterTemplate]

            helper_1(  # due to 1a), gets one more
                (len(test_parameters) - test_params_count),
                1,
                len(os.listdir(test_parameters_destination)) - test_params_count,
                1,
            )

            # dummy asserts
            dummy_processes = stores_config.all_template_stores[
                "dummy"
            ]._object_templates[ProcessTemplate]
            dummy_processes_destination = stores_config.all_template_stores[
                "dummy"
            ].store_folders[ProcessTemplate]
            helper_1(  # due to 1a), gets one more
                len(dummy_processes),
                3,
                len(os.listdir(dummy_processes_destination)),
                3,
            )
            dummy_parameters = stores_config.all_template_stores[
                "dummy"
            ]._attribute_templates[ParameterTemplate]
            dummy_parameters_destination = stores_config.all_template_stores[
                "dummy"
            ].store_folders[ParameterTemplate]
            helper_1(  # due to 1a),  gets one more too!
                len(dummy_parameters),
                2,
                len(os.listdir(dummy_parameters_destination)),
                2,
            )

        reusable_tests()

        ##### 1) b) same as above with already existing object template in store
        with self.assertWarns(ResourceWarning):
            p = Process(
                "dummy process 1",
                template=ProcessTemplate(
                    "dummy process template 3",  # same process
                    parameters=[
                        ParameterTemplate(
                            "dummy parameter template 4",
                            bounds=RealBounds(0, 10, "m"),  # new param
                        )
                    ],
                ),
            )

        test_processes = stores_config.all_template_stores["test"]._object_templates[
            ProcessTemplate
        ]
        processes_destination = stores_config.all_template_stores["test"].store_folders[
            ProcessTemplate
        ]
        # no addition at all for both tests and dummy
        #  why? template already exists and is reused with the attributes that come with it!
        # 1a) and 1b) have the same exepcted results
        reusable_tests()

        ##### 1) c) same as above with already existing attr template in store
        p = Process(
            "dummy process 1",
            template=ProcessTemplate(
                "dummy process template 4",  # new process
                parameters=[
                    ParameterTemplate(
                        "dummy parameter template 2",  # already existing param
                        bounds=RealBounds(0, 10, "m"),
                    )
                ],
            ),
        )

        test_processes = stores_config.all_template_stores["test"]._object_templates[
            ProcessTemplate
        ]
        processes_destination = stores_config.all_template_stores["test"].store_folders[
            ProcessTemplate
        ]
        # this time 1c) has diff exepected results
        # both dictionary gets one more process template... but same parameter templates!
        helper_1(
            len(test_processes) - test_processes_count,
            1 + 1,
            len(os.listdir(processes_destination)) - test_processes_count,
            1 + 1,
        )
        test_parameters = stores_config.all_template_stores[
            "test"
        ]._attribute_templates[ParameterTemplate]
        parameters_destination = stores_config.all_template_stores[
            "test"
        ].store_folders[ParameterTemplate]
        helper_1(
            (len(test_parameters) - test_params_count),
            1,
            len(os.listdir(parameters_destination)) - test_params_count,
            1,
        )

        dummy_processes = stores_config.all_template_stores["dummy"]._object_templates[
            ProcessTemplate
        ]
        dummy_processes_destination = stores_config.all_template_stores[
            "dummy"
        ].store_folders[ProcessTemplate]
        helper_1(
            len(dummy_processes),
            4,
            len(os.listdir(dummy_processes_destination)),
            4,
        )
        dummy_parameters = stores_config.all_template_stores[
            "dummy"
        ]._attribute_templates[ParameterTemplate]
        dummy_parameters_destination = stores_config.all_template_stores[
            "dummy"
        ].store_folders[ParameterTemplate]
        helper_1(
            len(dummy_parameters),
            2,
            len(os.listdir(dummy_parameters_destination)),
            2,
        )

    # def test_multi_stores(self):
    # TODO: do tests with 2 diff orders, check which one gets pulled from

    # TODO: test stores_config.all_template_stores["dummy"].register_all_templates_from_files()

    # TODO: tests the n's, the from_ mroe extensively,



import uuid
from gemd.util.impl import recursive_foreach


def assign_uuid(obj, scope):
    """
    Recursively assign a uuid to every BaseEntity that doesn't already contain a uuid.

    This ensures that all of the pointers in the object can be replaced with LinkByUID objects

    Parameters
    ----------
    obj: BaseEntity
        object to recursively assign uuids to
    scope: str
        scope of the uuid to assign

    Returns
    -------
    None

    """

    def func(base_obj):
        if not (scope in base_obj.uids.keys()):
            #     base_obj.uids[scope] = {}
            # if len(base_obj.uids[scope]) == 0:
            base_obj.add_uid(scope, str(uuid.uuid4()))
        return

    recursive_foreach(obj, func)
    return



import ipywidgets
from IPython.display import JSON, display
import pandas as pd
import traitlets
import json
import ipydatagrid

class BODisplay(traitlets.HasTraits):
    df = traitlets.Instance(pd.DataFrame)

    def _ipython_display_(self):
        dg = ipydatagrid.DataGrid(self.df, selection_mode="row")
        out = ipywidgets.Output()

        def on_change(change):
            new_row = change["new"][-1]["r1"]
            j = json.loads(self.df.iloc[new_row]["context"])
            # print(j)
            # print(type(j))
            out.clear_output()
            with out:
                # display(JSON(j, expanded=True))
                display(j)

        dg.observe(on_change, ["selections"])
        gsl = ipywidgets.GridspecLayout(1, 2)
        gsl[0, 0] = dg
        gsl[0, 1] = out
        display(gsl)



"""
A class for anything that should have associated command line arguments
if anything extending it also extends Runnable
"""

# imports
from abc import ABC, abstractmethod


class HasArguments(ABC):
    """
    A small utility class for anything that should have associated command line arguments
    """

    @classmethod
    @abstractmethod
    def get_command_line_arguments(cls):
        """
        Get the list of argument names and the dictionary of argument names/default values
        to add to the argument parser

        :return: args, a list of argument names recognized by the argument parser
        :rtype: list(str)
        :return: kwargs, a dictionary of default argument values keyed by argument names
            recognized by the argument parser
        :rtype: dict
        """
        return [], {}



from typing import ClassVar

from gemd import MeasurementTemplate, PropertyTemplate, RealBounds, NominalReal

from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Weighting"]


class Weighting(Measurement):
    """Class representing weighting as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="Weighting", description="The measurement of weight"
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Weighed Mass",
            bounds=RealBounds(0, 150, "g"),
            description="Weighted Mass",
        ),
        default_value=NominalReal(0.0, "g"),
    )

    define_attribute(
        _ATTRS,
        template=PropertyTemplate(
            name="Mass Loss",
            bounds=RealBounds(0, 50, "g"),
            description="Relative mass loss",
        ),
        default_value=NominalReal(0.0, "g"),
    )

    finalize_template(_ATTRS, TEMPLATE)



'''Support for array attributes.'''

from typing import TypeAlias, ClassVar, Union, Optional, Any
import re

import numpy as np
import pandas as pd

from gemd import (
    MeasurementTemplate, MeasurementSpec, MeasurementRun, MaterialRun,
    Condition, Parameter, Property, NominalInteger, NominalReal, FileLink
)
from gemd.entity.template.attribute_template import AttributeTemplate
from gemd.entity.attribute.base_attribute import BaseAttribute

from .base_special_attrs import BaseSpecialAttrs
from .utils import AttrTypes, attr_template_dict, generate_tags, generate_source

__all__ = ['ArrAttrs']

Measurement: TypeAlias = Union[MeasurementSpec, MeasurementRun]

class ArrAttrs(BaseSpecialAttrs):
    '''
    Represent a single measurement's array attributes as ``DataFrame``s of attributes.

    Supports multiple arrays, each potentially with single attributes that
    apply to the entire array and indexed attributes with multiple values.

    If there are multiple arrays, each array must have the same attributes
    and the same length.
    '''

    ARR: ClassVar[str] = '_arr'
    IDX: ClassVar[str] = '_idx'
    RE: ClassVar[re.Pattern] = re.compile(f'^(.+){ARR}(\\d+)(?:{IDX}(\\d+))?$')

    def __init__(self, measurement: Measurement) -> None:
        '''
        Parse a measurement's array attributes and create ``DataFrame``s
        for the conditions, parameters, and properties.

        Array attributes are expected to be named as ``<base_name>_arr<array>``
        or ``<base_name>_arr<array>_idx<index>``.

        Parameters
        ----------
        measurement: MeasurementSpec or MeasurementRun
            Measurement containing appropriately-named array attributes.
        '''        

        self._measurement = measurement

        # single_attributes are those with only _arr
        # multiple_attributes are those with both _arr and _idx

        conditions_tuple = self._parse(measurement.conditions)
        self._single_conds: pd.DataFrame = conditions_tuple[0]
        self._multiple_conds: pd.DataFrame = conditions_tuple[1]
        self._other_conds: list[Condition] = conditions_tuple[2]

        parameters_tuple = self._parse(measurement.parameters)
        self._single_params: pd.DataFrame = parameters_tuple[0]
        self._multiple_params: pd.DataFrame = parameters_tuple[1]
        self._other_params: list[Parameter] = parameters_tuple[2]

        self._single_props: pd.DataFrame = pd.DataFrame()
        self._multiple_props: pd.DataFrame = pd.DataFrame()
        self._other_props: list[Property] = []

        if isinstance(measurement, MeasurementRun):
            properties_tuple = self._parse(measurement.properties)
            self._single_props = properties_tuple[0]
            self._multiple_props = properties_tuple[1]
            self._other_props = properties_tuple[2]

        self._reindex_all_dfs()

    @property
    def measurement(self) -> Measurement:
        '''The measurement containing these attributes.'''
        return self._measurement

    @property
    def single_conditions(self) -> pd.DataFrame:
        '''Conditions that apply to entire arrays.'''
        return self._single_conds.copy()

    @property
    def multiple_conditions(self) -> pd.DataFrame:
        '''Conditions with multiple values per array.'''
        return self._multiple_conds.copy()

    @property
    def other_conditions(self) -> list[Condition]:
        '''Non-array conditions.'''
        return self._other_conds.copy()

    @property
    def single_parameters(self) -> pd.DataFrame:
        '''Parameters that apply to entire arrays.'''
        return self._single_params.copy()

    @property
    def multiple_parameters(self) -> pd.DataFrame:
        '''Parameters with multiple values per array.'''
        return self._multiple_params.copy()

    @property
    def other_parameters(self) -> list[Parameter]:
        '''Non-array parameters.'''
        return self._other_params.copy()

    @property
    def single_properties(self) -> pd.DataFrame:
        '''Properties that apply to entire arrays.'''
        return self._single_props.copy()

    @property
    def multiple_properties(self) -> pd.DataFrame:
        '''Properties with multiple values per array.'''
        return self._multiple_props.copy()

    @property
    def other_properties(self) -> list[Property]:
        '''Non-array properties.'''
        return self._other_props.copy()

    def concat(
        self,
        single_df: Optional[pd.DataFrame] = None,
        multiple_df: Optional[pd.DataFrame] = None
        ) -> None:
        '''
        Concatenate data with the single and/or multiple attributes.

        Indices in the new data are respected and will replace rows with
        identical indices in the existing data.

        Parameters
        ----------
        single_df, multiple_df: DataFrame, optional
            Should contain the appropriate attribute columns to update the
            single attributes and multiple attributes, respectively.
            At least one of these must be provided.
        '''

        if single_df is None and multiple_df is None:
            raise ValueError('At least on of single_df and multiple_df must not be None.')
        
        if single_df is not None:

            single_attrs = pd.concat(self._single_df_list(), axis=1)
            single_attrs = pd.concat((single_attrs, single_df), axis=0)
            single_attrs = single_attrs[~single_attrs.index.duplicated(keep='last')]

            self._single_conds = single_attrs.loc[:, self._single_conds.columns]
            self._single_params = single_attrs.loc[:, self._single_params.columns]
            self._single_props = single_attrs.loc[:, self._single_props.columns]

        if multiple_df is not None:

            multiple_attrs = pd.concat(self._multiple_df_list(), axis=1)
            multiple_attrs = pd.concat((multiple_attrs, multiple_df), axis=0,)
            multiple_attrs = multiple_attrs[~multiple_attrs.index.duplicated(keep='last')]

            self._multiple_conds = multiple_attrs.loc[:, self._multiple_conds.columns]
            self._multiple_params = multiple_attrs.loc[:, self._multiple_params.columns]
            self._multiple_props = multiple_attrs.loc[:, self._multiple_props.columns]

        self._reindex_all_dfs()
        self._reindex_all_attrs()

    def remove(self, index: int) -> None:
        '''Remove a row from the attributes by index, then reindex.'''

        new_index = np.arange(len(self._single_conds)-1)

        for df in self._single_df_list():
            df.drop(index=index, inplace=True)
            df.index = new_index

        for df in self._multiple_df_list():
            df.drop(index=index, inplace=True)
            new_codes = np.array(df.index.codes)
            new_codes[0][new_codes[0] > index] -= 1
            df.index = pd.MultiIndex.from_arrays(new_codes)

        self._reindex_all_attrs()

    def update_object(self) -> None:
        '''
        Update the ``MeasurementSpec`` or ``MeasurementRun`` used to initialize
        with the attributes.

        This is performed in-place on the existing measurement and will
        overwrite any attributes added after ``ArrAttrs`` initialization.
        '''

        single_conds = self._extract_attrs_from_df(self._single_conds)
        multiple_conds = self._extract_attrs_from_df(self._multiple_conds)
        single_params = self._extract_attrs_from_df(self._single_params)
        multiple_params = self._extract_attrs_from_df(self._multiple_params)

        self._measurement.conditions = self._other_conds + single_conds + multiple_conds
        self._measurement.parameters = self._other_params + single_params + multiple_params

        if isinstance(self._measurement, MeasurementRun):
            single_props = self._extract_attrs_from_df(self._single_props)
            multiple_props = self._extract_attrs_from_df(self._multiple_props)
            self._measurement.properties = self._other_props + single_props + multiple_props

    @classmethod
    def object_from_file(
        cls,
        filename: str,
        url: str,
        material_run: MaterialRun,
        measurement_template: MeasurementTemplate,
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str],
        arr_pos: int = 0,
        brand: Optional[str] = None,
        model: Optional[str] = None,
        instr_id: Optional[str] = None,
        email: Optional[str] = None,
        iso_date: Optional[str] = None,
        **read_csv_kwargs: Any
        ) -> tuple[MeasurementSpec, MeasurementRun]:
        '''
        Create an empty ``MeasurementSpec`` and filled ``MeasurementRun`` from a delimited file.

        Parameters
        ----------
        filename: str
            The name of the file.
        url: str
            The URL or file path leading to `filename`.
        material_run: str
            The relevant ``MaterialRun``.
        measurement_template: MeasurementTemplate
            The measurement template for this measurement.
        attr_types: list of Type[Condition], Type[Parameter], Type[Property]
            The type of each attribute in the data, in order.
        columns: list[str]
            The names of the data columns.
        names: list[str]
            The names of the attributes corresponding to an attribute in `measurement_template`.
        units: list[str or None]
            The units of the attribute values. str for NominalReal, None for NominalInteger.
        origins: list[str]
            The origins of the attribute values.
        arr_pos: int, default 0
            The integer array position to use in the generated attribute names.
        brand: str, optional
            The brand of the instrument used for the measurement.
        model: str, optional
            The model of the instrument used for the measurement.
            Ignored if `brand` is ``None``.
        instr_id: str, optional
            A further identifier for the instrument used, such as a number.
            Ignored if `brand` or `model` is ``None``.
        email: str, optional
            Email of the person who performed the measurement.
        iso_date: str, optional
            Date the measurement was performed, in ISO format.
        **read_csv_kwargs: Any
            Passed to ``pandas.read_csv()``.
        '''

        df = pd.read_csv(filename, **read_csv_kwargs)

        tags = generate_tags(brand, model, instr_id)

        source = generate_source(email, iso_date)

        template_dict = attr_template_dict(measurement_template)

        arr_pos = int(arr_pos)
        if arr_pos < 0:
            raise ValueError('arr_pos must be a non-negative integer.')

        if not (len(attr_types) == len(columns) == len(names) == len(units) == len(origins)):
            raise ValueError('Iterable inputs must be equal in length.')

        conditions, parameters, properties = cls._create_attrs_from_df(
            df, template_dict, attr_types, columns, names, units, origins, arr_pos
        )

        spec = MeasurementSpec(
            name=measurement_template.name,
            template=measurement_template,
            tags=tags
        )

        run = MeasurementRun(
            name=measurement_template.name,
            spec=spec,
            tags=tags,
            material=material_run,
            source=source,
            conditions=conditions,
            parameters=parameters,
            properties=properties,
            file_links=[FileLink(filename=filename, url=url)]
        )

        return spec, run

    @classmethod
    def _parse(
        cls,
        attrs: list[BaseAttribute]
        ) -> tuple[pd.DataFrame, pd.DataFrame, list[BaseAttribute]]:
        '''
        Match each attribute's name with regular expression.

        Base name determines the column. For single attributes (_arr only),
        value determines the row. For multiple attributes (_arr and _idx),
        array value determines the first row level in the row ``MultiIndex``
        and index value determines the second row level.
        '''

        single_df = pd.DataFrame()
        multiple_df = pd.DataFrame(index=pd.MultiIndex.from_arrays([[],[]]))
        other_attrs = []

        for attr in attrs:

            match = cls.RE.match(attr.name)

            if match is not None:
                cls._add_attr_to_df(attr, match, single_df, multiple_df)
            else:
                other_attrs.append(attr)

        reindexed_dfs = cls._reindex_df(single_df, multiple_df)

        return (*reindexed_dfs, other_attrs)

    @classmethod
    def _add_attr_to_df(
        cls,
        attr: BaseAttribute,
        match: re.Match[str],
        single_df: pd.DataFrame,
        multiple_df: pd.DataFrame
        ) -> None:
        '''Given a matching attribute, add it to `single_df` or `multiple_df` as appropriate.'''

        # regular expression groups
        match_groups = match.groups()
        base_name = match_groups[0]
        arr = int(match_groups[1])

        # _arr<array> case
        if match_groups[2] is None:
            df = single_df
            row_pos = arr
        # _arr<array>_idx<index> case
        else:
            df = multiple_df
            row_pos = (arr, int(match_groups[2]))
  
        # add column if not there
        if base_name not in df.columns:
            df[base_name] = np.nan

        # place attribute in appropriate df
        df.loc[row_pos, base_name] = attr

    @classmethod
    def _reindex_df(
        cls,
        single_df: pd.DataFrame,
        multiple_df: pd.DataFrame
        ) -> tuple[pd.DataFrame, pd.DataFrame]:
        '''Add in any missing integer indices and sort.'''

        # determine maximum arr from both dfs and maximum idx
        if len(single_df) > 0 and len(multiple_df) > 0:
            arr_max = max(single_df.index.max(), multiple_df.index.levels[0].max())
            idx_max = multiple_df.index.levels[1].max()
        elif len(single_df) > 0:
            arr_max = single_df.index.max()
        elif len(multiple_df) > 0:
            arr_max = multiple_df.index.levels[0].max()
            idx_max = multiple_df.index.levels[1].max()

        if len(single_df) > 0:
            single_df = single_df.reindex(np.arange(arr_max+1))

        if len(multiple_df) > 0:
            multiple_df = multiple_df.reindex(
                pd.MultiIndex.from_product((np.arange(arr_max+1), np.arange(idx_max+1)))
            )

        return single_df, multiple_df

    def _reindex_all_dfs(self) -> None:
        '''Ensure that each ``DataFrame`` has the same (maximum) number of arrays.'''

        if (
            len(self._single_conds) == len(self._single_params) == len(self._single_props) == 
            len(self._multiple_conds) == len(self._multiple_params) == len(self._multiple_props)==0
        ):
            return

        arr_max = int(np.nanmax([
            self._single_conds.index.max(),
            self._single_params.index.max(),
            self._single_props.index.max()
        ]))

        self._single_conds = self._single_conds.reindex(np.arange(arr_max+1))
        self._single_params = self._single_params.reindex(np.arange(arr_max+1))
        self._single_props = self._single_props.reindex(np.arange(arr_max+1))
        
        if len(self._multiple_conds) == len(self._multiple_params) == len(self._multiple_props)==0:
            return

        idx_max = 0
        
        if len(self._multiple_conds) > 0:
            idx_max = max(idx_max, self._multiple_conds.index.levels[1].max())

        if len(self._multiple_params) > 0:
            idx_max = max(idx_max, self._multiple_params.index.levels[1].max())

        if len(self._multiple_props) > 0:
            idx_max = max(idx_max, self._multiple_props.index.levels[1].max())

        self._multiple_conds = self._multiple_conds.reindex(
            pd.MultiIndex.from_product((np.arange(arr_max+1), np.arange(idx_max+1)))
        )
        self._multiple_params = self._multiple_params.reindex(
            pd.MultiIndex.from_product((np.arange(arr_max+1), np.arange(idx_max+1)))
        )
        self._multiple_props = self._multiple_props.reindex(
            pd.MultiIndex.from_product((np.arange(arr_max+1), np.arange(idx_max+1)))
        )

        idx_row_maxes=[
            max(
                self._multiple_conds.loc[arr_val, :].last_valid_index() or 0,
                self._multiple_params.loc[arr_val, :].last_valid_index() or 0,
                self._multiple_props.loc[arr_val, :].last_valid_index() or 0,
            )
            for arr_val in range(arr_max+1)
        ]

        to_drop = [(arr_val, idx_row_max) for i, arr_val in enumerate(range(arr_max+1)) for idx_row_max in range(idx_row_maxes[i]+1, idx_max+1)]

        self._multiple_conds.drop(index=to_drop, inplace=True)
        self._multiple_params.drop(index=to_drop, inplace=True)
        self._multiple_props.drop(index=to_drop, inplace=True)

    def _single_df_list(self) -> list[pd.DataFrame]:
        '''Return a list of the single attributes ``DataFrame``s.'''
        return [self._single_conds, self._single_params, self._single_props]

    def _multiple_df_list(self) -> list[pd.DataFrame]:
        '''Return a list of the multiple attributes ``DataFrame``s.'''
        return [self._multiple_conds, self._multiple_params, self._multiple_props]

    def _reindex_all_attrs(self) -> None:
        '''Rename all attributes according to where they are in their ``DataFrame``.'''

        for df in self._single_df_list():
            for name, series in df.items():
                for i, attr in series.items():
                    if not pd.isna(attr):
                        attr.name = f'{name}{self.ARR}{i}'

        for df in self._multiple_df_list():
            for name, series in df.items():
                for i, attr in series.items():
                    if not pd.isna(attr):
                        attr.name = f'{name}{self.ARR}{i[0]}{self.IDX}{i[1]}'

    @classmethod
    def _create_attrs_from_df(
        cls,
        df: pd.DataFrame,
        template_dict: dict[str, AttributeTemplate],
        attr_types: AttrTypes,
        columns: list[str],
        names: list[str],
        units: list[Union[str, None]],
        origins: list[str],
        arr_pos: int
        ) -> tuple[list[Condition], list[Parameter], list[Property]]:
        '''Create conditions, parameters, and properties from the `df` values.'''

        conditions = []
        parameters = []
        properties = []

        for a, c, n, u, o in zip(attr_types, columns, names, units, origins):
            if a is Condition:
                to_append = conditions
            elif a is Parameter:
                to_append = parameters
            elif a is Property:
                to_append = properties
            else:
                raise ValueError(
                    'attr_types must consist of Type[Condition], Type[Parameter], '
                    f'or Type[Property], but found {type(a)}.'
                )
            for i, val in enumerate(df.loc[:, c]):
                to_append.append(a(
                    name=f'{n}{cls.ARR}{arr_pos}{cls.IDX}{i}',
                    value=(NominalInteger(val) if u is None else NominalReal(val, u)),
                    template=template_dict[n],
                    origin=o
                ))

        return conditions, parameters, properties



from typing import ClassVar

from gemd import (
    MeasurementTemplate,
    PropertyTemplate,
    CategoricalBounds,
    NominalCategorical,
)
from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["SEM"]


class SEM(Measurement):
    """Class representing SEM as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="SEM", description="SEM Characterization"
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    finalize_template(_ATTRS, TEMPLATE)



"""Commonly-used types."""

from typing import TypedDict, Union, Optional, Literal, Any
from typing_extensions import TypeAlias
from gemd import (
    MaterialTemplate,
    ProcessTemplate,
    MeasurementTemplate,
    PropertyTemplate,
    ParameterTemplate,
    ConditionTemplate,
    MaterialSpec,
    ProcessSpec,
    MeasurementSpec,
    IngredientSpec,
    MaterialRun,
    ProcessRun,
    MeasurementRun,
    IngredientRun,
    Condition,
    Parameter,
    Property,
    PropertyAndConditions,
)
from gemd.enumeration import Origin

AttrTemplate: TypeAlias = Union[PropertyTemplate, ParameterTemplate, ConditionTemplate]

ObjTemplate: TypeAlias = Union[MaterialTemplate, ProcessTemplate, MeasurementTemplate]

Template: TypeAlias = Union[AttrTemplate, ObjTemplate]

Spec: TypeAlias = Union[MaterialSpec, ProcessSpec, MeasurementSpec, IngredientSpec]

Run: TypeAlias = Union[MaterialRun, ProcessRun, MeasurementRun, IngredientRun]

AllGEMD: TypeAlias = Union[Template, Spec, Run]

AllGEMDMinusAttr: TypeAlias = Union[ObjTemplate, Spec, Run]

SpecOrRun: TypeAlias = Union[Spec, Run]

SpecOrRunLiteral: TypeAlias = Literal["spec", "run", "both"]

Attributes: TypeAlias = Union[Condition, Parameter, Property, PropertyAndConditions]


# class TagsDict(TypedDict):
#     """Define dictionary representation of hierarchical tags."""

#     spec: dict[str, dict]
#     run: dict[str, dict]


# class FileLinksDict(TypedDict):
#     """Define dictionary containing file link strings for spec and run."""

#     spec: tuple[str, ...]
#     run: tuple[str, ...]


# class ProcessDict(TypedDict):
#     """Define dictionary containing process string for the spec and run."""

#     spec: str
#     run: str


# class ValueAndOrigin(TypedDict):
#     """Attribute value and origin as a dictionary."""

#     value: dict[str, Any]
#     origin: Origin


# class PropsAndCondsDict(TypedDict):
#     """Define dictionary output for material spec properties and conditions."""

#     property: ValueAndOrigin
#     conditions: dict[str, ValueAndOrigin]


# class ValueOriginDict(TypedDict):
#     """Define dictionary output for process and measurement spec/run conditions/parameters."""

#     spec: Optional[ValueAndOrigin]
#     run: Optional[ValueAndOrigin]



# imports
import os, csv, json, warnings
import shutil
from pathlib import Path
from typing import Union
from dataclasses import dataclass
from abc import ABC, abstractmethod

# from openmsimodel.stores.gemd_template_store import GEMDTemplateStore
# from openmsimodel.stores.common import GEMDTemplateStore, stores_config



from openmsimodel.workflow.workflow import Workflow
from openmsimodel.entity.base.ingredient import Ingredient
from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.entity.base.material import Material
from openmsimodel.entity.base.process import Process
from openmsimodel.entity.base.base_element import BaseElement
from openmsimodel.subworkflow.subworkflow import Subworkflow
from typing import ClassVar, Type, Optional
from openmsimodel.entity.base.helpers import from_spec_or_run
from openmsimodel.utilities.typing import Spec, Run


class ProcessBlock(Subworkflow):
    """
    ProcessBlock is a type of Subworkflow intended to represent consecutive BaseElements in the order of 'Ingredients', 'Process', 'Material', and 'Measurements'.
    It is the natural order of GEMD objects, and of our BaseElements object, which are essentially GEMD wrappers. It is a loose class and can omit some elements of the block.
    It can be a powerful way to manipulate, link, dump, etc, GEMD objects together, while Blocks themselves can be linked with one another, facilitating repeat
    elements, linking for wide (i.e., many ingredients, many measurements) or vertical (i.e., long sequence of BaseElements) workflow, etc.
    """

    def __init__(  # FIXME
        self,
        name: str,
        process: Optional[Process],
        workflow: Workflow = None,
        material: Optional[Material] = None,
        ingredients: Optional[
            dict
        ] = {},  # TODO: names have to be unique? will that be a problem?
        measurements: Optional[dict] = {},
        _type: str = None,
    ):  # FIXME
        """initialization of ProcessBlock.

        Args:
            name (str): Block name
            process (Optional[Process]): process of block.
            workflow (Workflow, optional): workflow that block belongs to. Defaults to None.
            material (Optional[Material], optional): material of block. Defaults to None.
            ingredients (Optional[list, dict], optional): Ingredients of block. Defaults to {}.
            measurements (Optional[list,dict], optional): Ingredients of block. Defaults to {}.

        Raises:
            TypeError: 'process' argument is not set.
            TypeError: ingredients must be of type "dict".
            TypeError: measurement must be of type "dict".
        """

        Subworkflow.__init__(self, name, workflow=workflow)
        # if process is None: #FIXME
        #     raise TypeError("'process' argument is not set. ")
        if (material and (not isinstance(material, Material))) or (
            process and not isinstance(process, Process)
        ):
            raise TypeError(
                f"Expected Material and Process; got '{type(material)}' and '{type(process)}' "
            )
        self.material = material
        self.process = process
        if type(ingredients) == dict:
            self.ingredients = ingredients
        elif type(ingredients) == list:
            self.ingredients = {}
            for i in ingredients:
                if not isinstance(i, Ingredient):
                    raise TypeError(f"Expected 'Ingredient'; got {type(i)}")
                if i.name in self.ingredients.keys():
                    raise NameError(
                        f"ingredients must have unique names. Found a duplicate: {i.name}"
                    )
                self.ingredients[i.name] = i
        else:
            raise TypeError(
                f"Expected ingredients to be a list or dict; got {type(ingredients)}"
            )

        if type(measurements) == dict:
            self.measurements = measurements
        elif type(measurements) == list:
            self.measurements = {}
            for m in measurements:
                if not isinstance(m, Measurement):
                    raise TypeError(f"Expected 'Measurement'; got {type(m)}")
                if m.name in self.measurements.keys():
                    raise NameError(
                        f"measurements must have unique names. Found a duplicate: {m.name}"
                    )
                self.measurements[m.name] = m
        else:
            raise TypeError(
                f"Expected measurements to be a list or dict; got {type(measurements)}"
            )
        self.type = _type

    def link_within(self):
        """this functions links the specs and runs of the BaseElements in the current block."""
        # link ingredients to process
        if self.ingredients and self.process:
            for name in self.ingredients.keys():
                self.ingredients[name].spec.process = self.process.spec
                self.ingredients[name].run.process = self.process.run
        else:
            print(f"no ingredients and process for block '{self.name}' were linked. ")
        # link process to material
        if self.material and self.process:
            self.material.spec.process = self.process.spec
            self.material.run.process = self.process.run
        else:
            print(f"no material and process for block '{self.name}' were linked. ")
        # link measurements to material
        if self.measurements and self.material:
            for name in self.measurements.keys():
                self.measurements[name].run.material = self.material.run
        # else:
        #     print(f"measurements and material for block {self.name} were not linked. ")

    def link_prior(
        self, prior_block: "ProcessBlock", ingredient_name_to_link: str
    ):  # TODO: change to 'current_ing_name'
        """links the prior block's material to current ingredient.

        Args:
            prior_block (Block): prior block containing the material to link
            ingredient_name_to_link (str): name of the ingredient in current block to link to prior material
        """
        linked = False
        for name in self.ingredients.keys():
            if self.ingredients[name].run.name == ingredient_name_to_link:
                self.ingredients[name].spec.material = prior_block.material.spec
                self.ingredients[name].run.material = prior_block.material.run
                linked = True
        if not linked:
            print(
                f"Current block '{self.name}' couldn't be linked to block '{prior_block.name}'"
            )

    def link_posterior(
        self, posterior_block: "ProcessBlock", ingredient_name_to_link: str
    ):
        """link the posterior block's ingredient to current material

        Args:
            posterior_block (Block): posterior block containing the ingredient to link
            ingredient_name_to_link (str): _description_
        """
        for name in posterior_block.ingredients.keys():
            if posterior_block.ingredients[name].run.name == ingredient_name_to_link:
                posterior_block.ingredients[name].spec.material = self.material.spec
                posterior_block.ingredients[name].run.material = self.material.run

    def return_all(self) -> list:
        _all = []
        for i in self.ingredients.values():
            _all.append(i)
        if self.process:
            _all.append(self.process)
        if self.material:
            _all.append(self.material)
        for m in self.measurements.values():
            _all.append(m)
        return _all

    def return_all_gemd(self) -> list:
        _all_gemd = []
        for obj in self.return_all():
            _all_gemd.extend(obj.return_all_gemd())
        return _all_gemd

    def add_ingredient(self, ingredient: Ingredient):
        """add ingredient to block"""
        if not isinstance(ingredient, Ingredient):
            raise TypeError(f"expected 'Ingredient'; got '{type(ingredient)}' ")
        if ingredient.name in self.ingredients.keys():
            raise NameError(
                f"Ingredients must have unique names. Found a duplicate: {ingredient.name}"
            )
        self.ingredients[ingredient.name] = ingredient

    def add_process(self, process: Process):
        """add process to block"""
        if not isinstance(process, Process):
            raise TypeError(f"expected 'Process'; got '{type(process)}' ")
        self.process = process

    def add_material(self, material: Material):
        """add material to block"""
        if not isinstance(material, Material):
            raise TypeError(f"expected 'Material'; got '{type(material)}' ")
        self.material = process

    def add_measurement(self, measurement: Measurement):
        """add measurement to block"""
        if not isinstance(measurement, Measurement):
            raise TypeError(f"expected 'Measurement'; got '{type(measurement)}' ")
        if measurement.name in self.measurements.keys():
            raise NameError(
                f"measurements must have unique names. Found a duplicate: {measurement.name}"
            )
        self.measurements[measurement.name] = measurement

    @classmethod
    def from_spec_or_run(
        cls,
        name: str,
        *,
        notes: Optional[str] = None,
        spec: Spec = None,
        run: Run = None,
    ):
        initial = from_spec_or_run(name=name, notes=notes, spec=spec, run=run)
        # if isinstance(initial, Material):

    # @classmethod
    # def from_(self, any):
    #     if isinstance_base_element(any):
    #         if isinstance(any, Material):
    #             self.material = any
    #         if isinstance(any, Process):
    #             self.process = any
    #         elif isinstance(any, Ingredient):
    #             self.ingredients[any.name] = any
    #         elif isinstance(any, Measurement):
    #             self.measurements[any.name] = any
    #         self.link_within()
    # elif isinstance_all_gemd(any):
    #     if isinstance(any, Material):



# imports
import os, csv, json, warnings
import shutil
from pathlib import Path
from typing import Union
from dataclasses import dataclass
from abc import ABC, abstractmethod

from gemd.util.impl import set_uuids
from gemd.json import GEMDJson
from gemd.entity.template import PropertyTemplate, ParameterTemplate, ConditionTemplate
from gemd.entity.template import MaterialTemplate, MeasurementTemplate, ProcessTemplate
from openmsimodel.utilities.typing import Template
from openmsimodel.utilities.cached_isinstance_functions import (
    isinstance_template,
    isinstance_attribute_template,
    isinstance_object_template,
)
from openmsimodel.entity.impl import assign_uuid
from openmsimodel.utilities.logging import Logger


# from openmsimodel.stores.common import GEMDTemplateStore, stores_config
class StoresConfig:
    """
    Managing the default configurations for activating store, name of the default store and a list of ids of all template stores

    Example:
    >>> from openmsimodel.stores.gemd_template_store import StoresConfig, stores_config
    >>> stores_config = StoresConfig(activated=True) # default name is local
    >>> stores_config.designated_store_id
    "local"
    >>> stores_config = StoresConfig(activated=True, designated_store_id="local_2")

    """

    all_template_stores: dict
    activated: bool
    designated_store_id: str
    designated_root: str

    def __init__(
        self,
        activated=False,
        designated_store_id: str = "local",
        designated_root: str = Path(__file__) / "stores/local",
    ):
        self.all_template_stores = {}
        self.activated = activated
        if self.activated:
            self.designated_store_id = designated_store_id
            self.designated_root = designated_root
            self.deploy_store(designated_store_id)

    def deploy_store(self, name):
        if name in self.all_template_stores.keys():
            raise NameError(f"template store with id {name} already exists.")
        self.all_template_stores[name] = GEMDTemplateStore(name, load_all_files=False)
        self.all_template_stores[name].root = self.designated_root
        self.all_template_stores[name].initialize_store()

    def register_store(self, store):
        if store.id in self.all_template_stores.keys():
            raise NameError(f"template store with id {store.name} already exists.")
        if self.activated:
            self.all_template_stores[store.id] = store
            self.all_template_stores[store.id].initialize_store()


stores_config = StoresConfig()

# TODO: chck types of errors returned


# TODO: maybe add file path, + from subclass?
@dataclass
class GEMDTemplate:  # TODO: move to typing
    template: Union[
        PropertyTemplate,
        ParameterTemplate,
        ConditionTemplate,
        MaterialTemplate,
        MeasurementTemplate,
        ProcessTemplate,
    ]
    from_file: bool
    from_store: bool
    from_memory: bool
    from_subclass: bool


# TODO: test if not root assigned


class GEMDTemplateStore(ABC):
    """
    A class to hold and work with a set of GEMD template objects. Allows easier loading from
    a template_type_root of json dump files coupled with one or more dictionaries of static, hard-coded templates
    """

    # _root = Path(__file__).parent / "stores/global"

    def __init__(self, id, encoder=GEMDJson(), load_all_files=False):
        """
        encoder = a pre-created GEMD JSON encoder (optional)
        """

        self.id = id
        self.encoder = encoder  # TODO: separate from workflow one
        # self.logger = Logger()
        self._n_from_files = 0
        self._n_hardcoded = 0
        self._property_templates = {}
        self._parameter_templates = {}
        self._condition_templates = {}
        self._attribute_templates = {  # TODO: changes names to match with spec store
            PropertyTemplate: self._property_templates,
            ParameterTemplate: self._parameter_templates,
            ConditionTemplate: self._condition_templates,
        }
        self._material_templates = {}
        self._measurement_templates = {}
        self._process_templates = {}
        self._object_templates = {
            MaterialTemplate: self._material_templates,
            MeasurementTemplate: self._measurement_templates,
            ProcessTemplate: self._process_templates,
        }
        if load_all_files:
            self.register_all_templates_from_store()
        if stores_config.activated:
            stores_config.register_store(self)

    @property
    def root(self):
        return self._root

    @root.setter
    def root(self, path: str):
        # TODO raise error if not exists?
        self._root = path

    @property
    def registry_path(self):
        return self._root / "registry.csv"

    @property
    def registry_columns(self):
        return ["persistent_id", "name", "type"]

    @property
    def store_folders(self):
        return {
            PropertyTemplate: self._root / "property_templates",
            ConditionTemplate: self._root / "condition_templates",
            ParameterTemplate: self._root / "parameter_templates",
            MaterialTemplate: self._root / "material_templates",
            MeasurementTemplate: self._root / "measurement_templates",
            ProcessTemplate: self._root / "process_templates",
        }

    TempStringToTemp = {  # TODO: move to typing
        "property_template": PropertyTemplate,
        "condition_template": ConditionTemplate,
        "parameter_template": ParameterTemplate,
        "measurement_template": MeasurementTemplate,
        "material_template": MaterialTemplate,
        "process_template": ProcessTemplate,
    }

    def register_all_templates_from_store(self):
        """
        loads all templates supposed to reside as files from store folder
        """
        # TODO: use the registered hardtemplates
        for template_type in self._object_templates.keys():
            template_type_root = self.store_folders[template_type]
            for template_name in os.listdir(template_type_root):
                # TODO: maybe read the path and pass it
                with open(
                    os.path.join(template_type_root, template_name), "r"
                ) as template_file:  # TODO: change to regular load
                    template = json.dumps((json.load(template_file)))
                    template = self.encoder.raw_loads(template)
                self.register_new_template(template, from_file=True, from_store=True)

    def register_new_template(
        self,
        template: Template,
        from_file=False,
        from_store=False,
        from_memory=False,
        from_subclass=False,
    ):
        """
        Add a new template that's been read from a file
        """
        if not isinstance_template(template):
            errmsg = f' : Excepted an AttributeTemplate or BaseTemplate but got "{type(template).__name__}"!'
            raise TypeError(errmsg)
        name = template.name

        if self.encoder.scope not in template.uids.keys():
            assign_uuid(
                template, "auto"
            )  # TODO: fix this, has nothing to do with encoder. just change to if "auto"?

        dict_to_add_to = None
        if isinstance_attribute_template(
            template
        ):  # TODO: use instance or use the existence of a type key, which would make it conform but loss actual type specificity
            dict_to_add_to = self._attribute_templates[
                self.TempStringToTemp[template.typ]
            ]
        elif isinstance_object_template(template):
            dict_to_add_to = self._object_templates[self.TempStringToTemp[template.typ]]

        if dict_to_add_to is None:
            raise RuntimeError(f"ERROR: unrecognized template type {type(template)}!")

        if (
            name in dict_to_add_to.keys()
        ):  # TODO: return a message if name is similar to one in bank
            warning_msg = f"WARNING: template with name '{name}' already exists in store and will be used. "
            warnings.warn(
                warning_msg,
                ResourceWarning,
            )
            if (
                self.encoder.scope not in dict_to_add_to[name].template.uids
            ):  # FIXME: is this to handle in case any kind of files comes in?
                warning_msg = f"WARNING: template with name '{name}' found in store doesn't have {self.encoder.scope} uid."
                warnings.warn(
                    warning_msg,
                    ResourceWarning,
                )
            dict_to_add_to[name].from_memory = False
            dict_to_add_to[name].from_subclass = False
            dict_to_add_to[name].from_store = True
            return dict_to_add_to[name]

        if from_store:  # TODO: change to from_store
            # dict_to_add_to[name] = GEMDTemplate(template, True)
            self._n_from_files += 1  # TODO: keep this incr but add to __n for all, and increment differently
        else:
            # TODO: this is always going to be false if not in dict. just compare without name
            # finding template by full match
            # has_matching_template = False
            # ordered_template = ordered(
            #     template
            # )  # TODO: maybe remove persistent_id key here from ordered_template/ or not necessary based on previous code block?
            # # TODO: use hash comparison instead?
            # for obj in dict_to_add_to.values():
            #     if ordered(obj) == ordered_template:
            #         has_matching_template = True
            #         matching_template = obj
            #         self.logger.info(
            #             f"template '{name}' sucessfully found in store by matching its JSON content to a template in store. "
            #         )
            #         return matching_template

            # writing to registry
            # TODO: convert to dataframe manipulation to be safer
            with open(self.registry_path, "r+") as registry_csv_file:
                persistent_id = len(registry_csv_file.readlines()) - 1
                template.add_uid(
                    "persistent_id", persistent_id
                )  # TODO: maybe make sure it's not there?
                registry_writer = csv.writer(registry_csv_file, delimiter=",")
                registry_writer.writerow([persistent_id, name, template.typ])

            # writing template to file
            with open(
                self.store_folders[type(template)] / f"{name}_pid_{persistent_id}.json",
                "w",  # TODO: maybe use diff encoding of path
            ) as template_file:
                template_file.write(self.encoder.thin_dumps(template, indent=3))

        dict_to_add_to[name] = GEMDTemplate(
            template, from_file, from_store, from_memory, from_subclass
        )

        return dict_to_add_to[
            name
        ]  # TODO: maybe return full template and keep info about the template, like from_store, path, etc

    # @classmethod
    # @abstractmethod ?
    # shutil can cause problems down the road? permissions, etc
    def initialize_store(self):
        # TODO raise error if not exists
        if os.path.isdir(self.root):
            shutil.rmtree(self.root)
        os.mkdir(self.root)
        for subfolder in self.store_folders.values():
            os.mkdir(subfolder)
        with open(self.registry_path, "w") as registry_csv_file:
            registry_writer = csv.writer(registry_csv_file, delimiter=",")
            registry_writer.writerow(self.registry_columns)

    def add_missing_hardcoded_templates(self, attr_hardcoded, obj_hardcoded):
        """
        Add any templates in the hardcoded dictionary that haven't already been added from a deserialization step

        attr_hardcoded = a dictionary of hard-coded attribute templates, some of which may have been
            used previously and dumped as json into dirpath
        obj_hardcoded = a dictionary of hard-coded object templates, some of which may have been used
            previously and dumped as json into dirpath
        """
        self.__add_missing_templates(attr_hardcoded)
        self.__add_missing_templates(obj_hardcoded)

    def attr(self, template_name):
        """
        Return an attribute template given its name
        """
        try:
            return self._attribute_templates[template_name].template
        except KeyError:
            raise ValueError(
                f'ERROR: no stored attribute template called "{template_name}"'
            )

    def obj(self, template_name):
        """
        Return an object template given its name
        """
        try:
            for template_type in self._object_templates.keys():
                return self._object_templates[template_type][template_name].template
        except KeyError:
            raise ValueError(
                f'ERROR: no stored object template called "{template_name}"'
            )

    def __add_missing_templates(self, hardcoded):
        names_seen = set()
        for name, template in hardcoded.items():
            if name in names_seen:
                raise ValueError(
                    f"ERROR: harcoded template dictionary duplicates template name {name}!"
                )
            else:
                names_seen.add(name)
            if template.uids is not None and self.encoder.scope in template.uids.keys():
                errmsg = f'ERROR: "{self.encoder.scope}" scope UID has already been set for a hard-coded template!'
                raise RuntimeError(errmsg)
            dict_to_add_to = None
            if isinstance_attribute_template(template):
                dict_to_add_to = self._attribute_templates[
                    self.TempStringToTemp[template["type"]]
                ]
            elif isinstance_object_template(template):
                dict_to_add_to = self._object_templates[
                    self.TempStringToTemp[template["type"]]
                ]
            if dict_to_add_to is None:
                raise RuntimeError(
                    f"ERROR: unrecognized template type {type(template)}!"
                )
            if name in dict_to_add_to.keys():
                continue
            set_uuids(template, self.encoder.scope)
            dict_to_add_to[name] = GEMDTemplate(template, False)
            self._n_hardcoded += 1

    @property
    def n_from_files(self):
        return self._n_from_files

    @property
    def n_hardcoded(self):
        return self._n_hardcoded

    @property
    def all_templates(self):
        both_dicts = [self._attribute_templates, self._object_templates]
        for tempdict in both_dicts:
            for template_type in tempdict.keys():
                for name in tempdict[template_type].keys():
                    yield tempdict[template_type][name].template

    @property
    def all_read_templates(self):
        both_dicts = [self._attribute_templates, self._object_templates]
        for tempdict in both_dicts:
            for template_type in tempdict.keys():
                for name in tempdict[template_type].keys():
                    if tempdict[template_type][name].from_store:
                        yield tempdict[template_type][name].template



from openmsimodel.workflow.workflow import Workflow
from openmsimodel.workflow.folder_or_file import FolderOrFile
from openmsimodel.subworkflow.process_block import ProcessBlock
from openmsimodel.entity.base import Material, Process, Measurement, Ingredient

from openmsimodel.entity.processes.birdshot.summarize import (
    Summarize,
)
from openmsimodel.entity.processes.birdshot.infer_compositions import InferCompositions
from openmsimodel.entity.processes.birdshot.select_composition import SelectComposition
from openmsimodel.entity.processes.birdshot.add_material import AddMaterial
from openmsimodel.entity.processes.birdshot.mixing import Mixing
from openmsimodel.entity.processes.birdshot.arc_melting import ArcMelting
from openmsimodel.entity.processes.birdshot.homogenization import Homogenization
from openmsimodel.entity.processes.birdshot.forging import Forging
from openmsimodel.entity.processes.birdshot.setting_traveler import SettingTraveler
from openmsimodel.entity.processes.birdshot.setting_traveler_sample import (
    SettingTravelerSample,
)
from openmsimodel.entity.processes.birdshot.aggregate_traveler_samples import (
    AggregateTravelerSamples,
)


from openmsimodel.entity.materials.birdshot.summary import Summary
from openmsimodel.entity.materials.birdshot.inferred_alloy_compositions import (
    InferredAlloyCompositions,
)
from openmsimodel.entity.materials.birdshot.composition import Composition
from openmsimodel.entity.materials.birdshot.element import Element
from openmsimodel.entity.materials.birdshot.alloy import Alloy
from openmsimodel.entity.materials.birdshot.traveler import Traveler
from openmsimodel.entity.materials.birdshot.traveler_sample import TravelerSample
from openmsimodel.entity.materials.birdshot.traveler_samples import TravelerSamples

from openmsimodel.entity.measurements.birdshot.weighting import Weighting
from openmsimodel.entity.measurements.birdshot.measure_dimensions import (
    MeasureDimensions,
)
from openmsimodel.entity.measurements.birdshot.sem import SEM
from openmsimodel.entity.measurements.birdshot.ni import NI
from openmsimodel.entity.measurements.birdshot.xrd import XRD
from openmsimodel.entity.measurements.birdshot.tensile import Tensile
from openmsimodel.entity.measurements.birdshot.mounting_and_polishing import (
    MountingAndPolishing,
)
from openmsimodel.entity.measurements.birdshot.srjt import SRJT

from openmsimodel.utilities.tools import plot_graph
from openmsimodel.utilities.argument_parsing import OpenMSIModelParser

import json
import os
import shutil
import pandas as pd
from pathlib import Path
from collections import defaultdict
import time
import sys

import subprocess

from gemd.entity.object import MaterialRun, ProcessRun, IngredientRun, MeasurementRun
from gemd.entity.attribute import Property, Parameter, Condition, PropertyAndConditions
from gemd.entity.value import NominalReal, NominalCategorical
from gemd.entity.util import make_instance
from gemd.entity import PerformedSource, FileLink
from gemd.json import GEMDJson
from gemd.util.impl import recursive_foreach

# helpers folder is specific to birdshot, and found in the birdshot gemd/helpers folder
# from helpers.attribute_templates import ATTR_TEMPL
# from helpers.object_templates import OBJ_TEMPL
# from helpers.object_specs import OBJ_SPECS

# TODO: write ingredients?
# TODO: add number/quantity attribute to SuggestedCompositions
# TODO: add ingredient attributes
# TODO: add aggregate/purchasing details
# TODO: change the _spec to _run? or not?
# TODO: add form to alloy object (melted, etc)
# TODO: figure out path offset automatically
# TODO: figure out functions for naming output files (fn)
# FIXME: in yfiles, Diffusion shows up with out " " as opposed to rest: fix


class BIRDSHOTWorfklow(Workflow, FolderOrFile):
    def __init__(self, root, output, iteration, synthesis_path, srjt_path):
        """
        this function represents a TAMU workflow, from a list of compositions to tests, through their
        VAM and DED fabrications, their splits into travelers for numerous characterizations, all the way to
        using the characterization results to infer the next compositions using BO.
        :param iteration: the iteration of the workflow (i.e., AAB)
        :param synthesis_path: the path to synthesis file (i.e., /Sample Data/Iteration2_AAB/HTMDEC AAB Summary Synthesis Results)
        :param srjt_path: the path to srjt file
        """
        Workflow.__init__(self)
        # self.root gets set in FolderOrFile
        FolderOrFile.__init__(self, root, parent_path=None, is_last=False)
        self.encoder = GEMDJson()
        self.output = Path(output)
        self.iteration = iteration
        self.synthesis_path = Path(synthesis_path)
        self.srjt_path = Path(srjt_path)
        recursive_dict = lambda: defaultdict(recursive_dict)
        self.subs = recursive_dict()  # overwrite
        self.terminal_blocks = recursive_dict()  # overwrite
        self.file_links = recursive_dict()
        self.measurements = recursive_dict()
        self.measurement_types = {
            "SEM": SEM,
            "XRD": XRD,
            "NI": NI,
            "Tensile": Tensile,
            "Mounting and Polishing ": MountingAndPolishing,
        }
        self.aggregate_or_buy = True
        self.testing_mode = False

    def build(self):
        # ingesting the results of the summary sheet, which is associated with each composition space, and workflow object
        ingest_synthesis_results(
            self.iteration, self.synthesis_path, self.file_links, self.measurements
        )

        ingest_srjt_results(self.srjt_path, self.output, self.measurements)

        ############## block 1: first infer_compositions_block
        def create_infer_compositions_block():
            # collecting all the compositions in the composition space to be assigned as tags
            compositions_tags = []
            for composition_id in self.measurements.keys():
                for element_name, element_percentage in self.measurements[
                    composition_id
                ]["Target Composition, at.%"].items():
                    compositions_tags.append(
                        (
                            "{}::{}".format(composition_id, element_name),
                            str(element_percentage),
                        )
                    )

            inferred_alloy_compositions = InferredAlloyCompositions(
                "Inferred Alloy Compositions"
            )
            inferred_alloy_compositions._set_tags(
                tags=compositions_tags,
                spec_or_run=inferred_alloy_compositions.run,
            )
            for obj in [
                inferred_alloy_compositions.run,
                inferred_alloy_compositions.spec,
            ]:
                inferred_alloy_compositions._set_filelinks(
                    spec_or_run=obj,
                    supplied_links={"Summary Sheet": self.file_links["Summary Sheet"]},
                )
            infer_compositions_block_name = "Infer Compositions"
            infer_compositions_block = ProcessBlock(
                name=infer_compositions_block_name,
                workflow=self,
                ingredients=[],
                process=None,
                material=inferred_alloy_compositions,
            )
            infer_compositions_block.link_within()
            self.subs[infer_compositions_block_name] = infer_compositions_block
            return infer_compositions_block, inferred_alloy_compositions

        (
            infer_compositions_block,
            inferred_alloy_compositions,
        ) = create_infer_compositions_block()

        self.subs[infer_compositions_block.name] = infer_compositions_block

        # extracting tree structure to build the model up
        count = 0
        path_offset = 6
        tree_folders_and_files = self.make_tree(FolderOrFile, self.root)

        # print(self.displayable())
        # with open(self.output / "folder_structure.txt", "w") as fp:
        #     for p in tree_folders_and_files:
        #         fp.write(p.displayable() + "\n")

        ############## blocks from 2 to (n-3)
        # looping through all the folders and files in the tree structure
        for item in tree_folders_and_files:
            item_path = str(item.root)
            if os.path.isfile(item_path):
                continue
            onlyfiles = [
                f
                for f in os.listdir(item_path)
                if os.path.isfile(os.path.join(item_path, f))
            ]
            # /AAA/VAM/B/AAA01/T01 or /AAA/DED/A/AAA01-AAA08/AAA01
            if item.depth >= 4:
                is_ded = "DED" in item_path
                split_item_path = substring_after(item_path, str(self.root))
                if split_item_path[0] == "/":  # avoiding first '/'
                    split_item_path = split_item_path[1:]
                split_item_path = split_item_path.split("/")
                fabrication_method = split_item_path[
                    0
                ]  # path offset used to removes path elements before data (i.e., ../data/)
                batch = split_item_path[1]
                if is_ded:
                    if item.depth == 4:
                        continue
                    elif item.depth >= 5:
                        composition_id = split_item_path[3]
                else:
                    composition_id = split_item_path[2]
                if (item.depth == 4 and not is_ded) or (item.depth == 5 and is_ded):
                    if onlyfiles:
                        self.process(
                            item_path,
                            composition_id,
                            batch,
                            fabrication_method,
                            inferred_alloy_compositions,
                            infer_compositions_block,
                        )
                # AAA/VAM/B/AAA01/T01/T02 or ( AAA/DED/A/AAA01-AAA08/AAA01/C01/EDS VS AAA/DED/B/AAA01-AAA08/AAA01/C01/G01 )
                if (
                    (item.depth >= 5 and not is_ded) or (item.depth >= 6 and is_ded)
                ) and (onlyfiles):
                    self.process_measurement(
                        item_path,
                        composition_id,
                        batch,
                        fabrication_method,
                    )

        # adding average tensile measurement manually
        def add_average_tensile_measurement():
            """
            assigns the average tensile measurement from the excel summary sheet to the model
            (since it is reflected only in the summary sheet, not in the tree structure as a folder/measurement or else)
            """
            for composition_id in self.terminal_blocks.keys():
                for fabrication_method in self.terminal_blocks[composition_id].keys():
                    for batch in self.terminal_blocks[composition_id][
                        fabrication_method
                    ].keys():
                        self.process_measurement(
                            "Average/Tensile",
                            composition_id,
                            batch,
                            fabrication_method,
                            bypass=True,
                        )

        add_average_tensile_measurement()

        ############## block -3: aggregation of traveler samples by category
        def extract_traveler_samples_blocks():
            traveler_samples_blocks = defaultdict(lambda: [])
            for composition_id in self.terminal_blocks.keys():
                for fabrication_method in self.terminal_blocks[composition_id].keys():
                    for batch in self.terminal_blocks[composition_id][
                        fabrication_method
                    ].keys():
                        terminal_blocks = self.terminal_blocks[composition_id][
                            fabrication_method
                        ][batch]
                        for terminal_block in terminal_blocks:
                            traveler_samples_blocks[terminal_block.type].append(
                                terminal_block
                            )
            return traveler_samples_blocks

        traveler_samples_blocks = extract_traveler_samples_blocks()

        terminal_blocks = []
        for traveler_sample_type in traveler_samples_blocks.keys():
            ings = []
            for traveler_sample_block in traveler_samples_blocks[traveler_sample_type]:
                ings.append(
                    Ingredient(f"{traveler_sample_block.material.run.name} Ing.")
                )
            aggregate_traveler_samples_process = AggregateTravelerSamples(
                f"Aggregating {traveler_sample_type} samples"
            )
            traveler_samples_material = TravelerSamples(
                f" {traveler_sample_type} Traveler samples"
            )

            aggregate_traveler_samples_block = ProcessBlock(
                name=f"Aggregating {traveler_sample_type} Traveler Samples",
                workflow=self,
                ingredients=ings,
                process=aggregate_traveler_samples_process,
                material=traveler_samples_material,
                _type="procedural",
            )
            aggregate_traveler_samples_block.link_within()
            for traveler_sample_block in traveler_samples_blocks[traveler_sample_type]:
                aggregate_traveler_samples_block.link_prior(
                    traveler_sample_block,
                    ingredient_name_to_link=f"{traveler_sample_block.material.run.name} Ing.",
                )
            terminal_blocks.append(aggregate_traveler_samples_block)

        ############## block -2: aggregation of summary sheet block
        aggregate_summary_process = Summarize("Summarize inputs to B.O")
        summary_material = Summary("Summary")
        summary_block = ProcessBlock(
            name="Summary",
            workflow=self,
            ingredients=[],
            process=aggregate_summary_process,
            material=summary_material,
            _type="manual",
        )

        def link_traveler_samples_to_summary_sheet():
            """
            this functions takes all the traveler samples build in the process_measurement() calls,
            and converts them to ingredients to the next process, which is aggregation the summary sheet block, a
            and which will ultimately be used for the Bayesian optimization process.
            """
            for terminal_block in terminal_blocks:
                ingredient_name = "{} Ing.".format(terminal_block.material._run.name)
                ingredient = Ingredient(ingredient_name)
                summary_block.ingredients[ingredient.name] = ingredient
                summary_block.link_within()
                summary_block.link_prior(
                    terminal_block, ingredient_name_to_link=ingredient_name
                )
            self.subs[summary_block.name] = summary_block
            # self.terminal_blocks[
            #     summary_block.name
            # ] = summary_block

        # an important step to link all the measurements on traveler samples to the summary sheet of the NEXT iteration
        link_traveler_samples_to_summary_sheet()

        ############## block -1: Bayesian Optimization block
        summary_sheet_ingredient_name = summary_block.material._run.name
        summary_sheet_ingredient = Ingredient(
            "{} Ing.".format(summary_sheet_ingredient_name)
        )
        infer_compositions_process = InferCompositions("Infer comp. using B.O.")
        infer_next_compositions_block = ProcessBlock(
            name="Infer Compositions",
            workflow=self,
            ingredients=[summary_sheet_ingredient],
            process=infer_compositions_process,
            material=None,
            _type="procedural",
        )

        infer_next_compositions_block.link_within()
        infer_next_compositions_block.link_prior(
            summary_block,
            ingredient_name_to_link=summary_sheet_ingredient._run.name,
        )
        self.subs[summary_block.name] = summary_block

        # assigning terminal material
        self.terminal_material = infer_next_compositions_block.process.run

    def process(
        self,
        item_path,
        composition_id,
        batch,
        fabrication_method,
        inferred_alloy_compositions,
        prior_block,
    ):
        fabrication_method_lowercase = fabrication_method.lower()
        yymm = None

        def read_details():
            try:
                processing_details = json.load(
                    open(
                        os.path.join(
                            item_path,
                            "{}-processing-details-v1.json".format(
                                fabrication_method_lowercase
                            ),
                        )
                    )
                )
                synthesis_details = json.load(
                    open(
                        os.path.join(
                            item_path,
                            "{}-synthesis-details-v1.json".format(
                                fabrication_method_lowercase
                            ),
                        )
                    )
                )
                traveler = json.load(
                    open(
                        os.path.join(
                            item_path,
                            "{}-traveler-v1.json".format(fabrication_method_lowercase),
                        )
                    )
                )
                yymm = traveler["data"]["Sample ID"]["Year & Month"]
                return processing_details, synthesis_details
            except:
                return {}, {}

        processing_details, synthesis_details = read_details()

        (
            long_name,
            short_name,
            alloy_common_name,
            common_tags,
        ) = return_common_items(composition_id, fabrication_method, batch, yymm)

        ############## block 1: Selecting composition
        inferred_alloy_compositions_ingredient_name = "{} Ing.".format(
            inferred_alloy_compositions._spec.name
        )
        inferred_alloy_compositions_ingredient = Ingredient(
            inferred_alloy_compositions_ingredient_name
        )
        select_composition_process = SelectComposition(f"Select {composition_id}")
        composition_material = Composition(f"Composition {composition_id}")

        def set_composition_material_and_select_composition_process_params_and_tags():
            composition_tags = tuple(
                [
                    (element_name, str(element_percentage))
                    for element_name, element_percentage in self.measurements[
                        composition_id
                    ]["Target Composition, at.%"].items()
                ]
            )
            composition_and_common_tags = common_tags + composition_tags
            select_composition_process._set_tags(
                tags=composition_and_common_tags,
                spec_or_run=select_composition_process.run,
            )
            composition_material._set_tags(
                tags=composition_and_common_tags,
                spec_or_run=composition_material.run,
            )
            item_path_file_link = FileLink(filename="root", url=item_path)
            select_composition_process._set_filelinks(
                spec_or_run=select_composition_process.run,
                supplied_links={"root": item_path_file_link},
            )
            composition_material._set_filelinks(
                spec_or_run=composition_material.run,
                supplied_links={"root": item_path_file_link},
            )
            for child in os.listdir(item_path):
                child_path = os.path.join(item_path, child)
                item_path_file_link = FileLink(filename=child, url=child_path)
                select_composition_process._set_filelinks(
                    spec_or_run=select_composition_process.run,
                    supplied_links={child: item_path_file_link},
                )
                composition_material._set_filelinks(
                    spec_or_run=composition_material.run,
                    supplied_links={child: item_path_file_link},
                )

                return composition_tags

        composition_tags = (
            set_composition_material_and_select_composition_process_params_and_tags()
        )

        block1 = ProcessBlock(
            name=f"Selecting Composition {composition_id}",
            workflow=self,
            ingredients=[inferred_alloy_compositions_ingredient],
            process=select_composition_process,
            material=composition_material,
        )
        block1.link_within()
        block1.link_prior(
            prior_block,
            ingredient_name_to_link=inferred_alloy_compositions_ingredient_name,
        )
        self.subs[block1.name] = block1

        ############## block 2: Weighting + aggregating (or buying) materials (one to many)
        composition_elements = []
        parallel_block2s = []
        composition_ingredient_name = "{} Ing.".format(composition_material._spec.name)
        # composition_ingredient = Ingredient(composition_ingredient_name)
        for (
            element_name,
            element_property,
        ) in composition_tags:  # TODO: if aggregate_or_buy
            composition_ingredient = Ingredient(composition_ingredient_name)
            adding_material_process = AddMaterial(
                f"Adding {element_name} for {composition_id}"
            )
            adding_material_process._set_tags(
                tags=common_tags,
                spec_or_run=adding_material_process.run,
            )
            composition_element_material = Element(
                "{} in {}".format(element_name, composition_id)
            )

            preparation_metadata = synthesis_details["data"]["Material Preparation"][""]
            weighting_performed_source = PerformedSource(
                preparation_metadata["Completed By"],
                preparation_metadata["Finish Date"],
            )

            def set_composition_element_material_params_and_tags():
                target_mass = synthesis_details["data"]["Material Preparation"][
                    "Target Mass"
                ][element_name]

                composition_element_material._update_attributes(
                    AttrType=PropertyAndConditions,
                    attributes=(
                        PropertyAndConditions(
                            property=Property(
                                "Composition Percentage",
                                value=NominalReal(float(element_property), ""),
                                origin="predicted",
                                notes="",
                                file_links=[],
                            ),
                            conditions=[],
                        ),
                        PropertyAndConditions(
                            property=Property(
                                "Target Mass",
                                value=NominalReal(float(target_mass), "g"),
                                origin="computed",
                                notes="",
                                file_links=[],
                            ),
                            conditions=[],
                        ),
                    ),
                    which="spec",
                )
                composition_element_material._set_tags(
                    tags=common_tags,
                    spec_or_run=composition_element_material.run,
                )
                composition_element_material._run._sample_type = "production"

            set_composition_element_material_params_and_tags()

            weighting_measurement = Weighting(
                "Weigh {} for {}".format(element_name, composition_id)
            )
            weighting_measurement._set_tags(
                tags=common_tags,
                spec_or_run=weighting_measurement.run,
            )
            weighted_mass = synthesis_details["data"]["Material Preparation"][
                "Weighed Mass"
            ][element_name]
            weighting_measurement._update_attributes(
                AttrType=Property,
                attributes=(
                    Property(
                        "Weighed Mass", value=NominalReal(float(weighted_mass), "g")
                    ),
                ),
                which="run",
            )
            weighting_measurement._run.source = weighting_performed_source

            composition_elements.append(composition_element_material)
            block2 = ProcessBlock(
                name="Aggregating {} for {}".format(element_name, short_name),
                workflow=self,
                ingredients=[composition_ingredient],
                process=adding_material_process,
                material=composition_element_material,
                measurements=[weighting_measurement],
                _type="procedural",
            )
            block2.link_within()
            block2.link_prior(
                block1, ingredient_name_to_link=composition_ingredient_name
            )
            parallel_block2s.append(block2)
            self.subs[block2.name] = block2

        ############## block 3 : mixing
        composition_element_ingredients = []
        for composition_element_material in composition_elements:
            composition_element_ingredient_name = "{} Ing.".format(
                composition_element_material._run.name
            )
            composition_element_ingredient = Ingredient(
                composition_element_ingredient_name
            )
            composition_element_ingredients.append(composition_element_ingredient)
        mixing_process = Mixing(f"Mix elements of {composition_id}")
        mixing_process._set_tags(
            tags=common_tags,
            spec_or_run=mixing_process.run,
        )
        alloy_material = Alloy(alloy_common_name)
        alloy_material._set_tags(
            tags=common_tags,
            spec_or_run=alloy_material.run,
        )
        srjt_measurement = SRJT(f"SRJT charact. for {composition_id}")
        srjt_measurement._set_tags(
            tags=common_tags,
            spec_or_run=srjt_measurement.run,
        )

        for srjt_property_name in self.measurements[composition_id]["SRJT"].keys():
            attribute_value = self.measurements[composition_id]["SRJT"][
                srjt_property_name
            ]
            # if not type(value) == str:  # FIXME
            unit = srjt_measurement._ATTRS["properties"][srjt_property_name][
                "obj"
            ].bounds.default_units
            value = NominalReal(float(attribute_value), unit)
            srjt_measurement._update_attributes(
                AttrType=Property,
                attributes=(
                    Property(
                        srjt_property_name,
                        value=value,
                        origin="measured",
                        notes="",
                        file_links=[],
                    ),
                ),
                which="run",
            )

        # assigning srjt measurement to alloy at earlier phase due to measuremenet being indep. from batch/fab method

        block3 = ProcessBlock(
            name="Mix elements of {}".format(composition_id),
            workflow=self,
            ingredients=composition_element_ingredients,
            process=mixing_process,
            material=alloy_material,
            measurements=[srjt_measurement],
        )
        block3.link_within()
        for i, block in enumerate(parallel_block2s):
            block3.link_prior(
                block,
                ingredient_name_to_link=composition_element_ingredients[i]._run.name,
            )
        self.subs[block3.name] = block3

        ############## block 4 : fabrication
        if fabrication_method == "VAM":
            self.process_vam(
                composition_id,
                fabrication_method,
                batch,
                synthesis_details,
                processing_details,
                common_tags,
                alloy_common_name,
                block3,
            )

    def process_vam(
        self,
        composition_id,
        fabrication_method,
        batch,
        synthesis_details,
        processing_details,
        common_tags,
        alloy_common_name,
        prior_block,
    ):
        ############## block 4
        melted_alloy_properties = synthesis_details["data"]["Arc Melting"]["   "]

        arc_melting_metadata = synthesis_details["data"]["Arc Melting"][" "]
        arc_melting_performed_source = PerformedSource(
            arc_melting_metadata["Completed By"],
            arc_melting_metadata["Finish Date"],
        )

        alloy_ingredient_name = "{} Ing.".format(alloy_common_name)
        alloy_ingredient = Ingredient(alloy_ingredient_name)

        arc_melting_process = ArcMelting("Arc melting {}".format(alloy_common_name))

        def gen_arc_melting_tags():
            arc_melting_tags = ()
            for name in ["3 Part Sections", "Full Ingot"]:
                _d = synthesis_details["data"]["Arc Melting"][name]
                for key, value in _d.items():
                    tag_name = "{}::{}".format(name, key)
                    arc_melting_tags = ((tag_name, str(value)),) + arc_melting_tags
            arc_melting_tags = (
                ("start_date", arc_melting_metadata["Start Date"]),
                ("finish_date", arc_melting_metadata["Finish Date"]),
                ("time_spent", str(arc_melting_metadata["Time Spent"])),
            ) + arc_melting_tags
            return arc_melting_tags

        arc_melting_tags = gen_arc_melting_tags()
        arc_melting_and_common_tags = common_tags + arc_melting_tags
        arc_melting_process._set_tags(
            tags=arc_melting_and_common_tags,
            spec_or_run=arc_melting_process.run,
        )

        def set_arc_melting_process_params():
            arc_melting_parameters = synthesis_details["data"]["Arc Melting"]["  "]
            for attribute_name, attribute_value in arc_melting_parameters.items():
                if type(attribute_value) == str:
                    value = NominalCategorical(attribute_value)
                else:
                    unit = ""
                    if attribute_name == "Initial Purging Times":
                        unit = "hour"
                    value = NominalReal(attribute_value, unit)
                arc_melting_process._update_attributes(
                    AttrType=Parameter,
                    attributes=(
                        Parameter(attribute_name, value=value, origin="specified"),
                    ),
                    which="run",
                )

        set_arc_melting_process_params()
        arc_melting_process._run.source = arc_melting_performed_source

        melted_alloy_material = Alloy("Arc Melted {}".format(alloy_common_name))
        melted_alloy_material._set_tags(
            tags=common_tags,
            spec_or_run=melted_alloy_material.run,
        )
        melted_alloy_material._update_attributes(
            AttrType=PropertyAndConditions,
            attributes=(
                PropertyAndConditions(
                    property=Property(
                        "Form",
                        value=NominalCategorical("Ingot"),
                        origin="predicted",
                        notes="",
                        file_links=[],
                    ),
                    conditions=[],
                ),
            ),
            which="spec",
        )

        weighting_alloy_measurement = Weighting(
            "Weigh {}".format(melted_alloy_material.name)
        )
        weighting_alloy_measurement._set_tags(
            tags=common_tags,
            spec_or_run=weighting_alloy_measurement.run,
        )
        weighting_alloy_measurement._update_attributes(
            AttrType=Property,
            attributes=(
                Property(
                    "Weighed Mass",
                    value=NominalReal(
                        float(melted_alloy_properties["Final Ingot Mass"]), "g"
                    ),
                ),
                Property(
                    "Mass Loss",
                    value=NominalReal(float(melted_alloy_properties["Mass Loss"]), "g"),
                ),
            ),
            which="run",
        )
        weighting_alloy_measurement._run.source = arc_melting_performed_source

        block4 = ProcessBlock(
            name=f"Arc Melting of Alloy {alloy_common_name}",
            workflow=self,
            ingredients=[alloy_ingredient],
            process=arc_melting_process,
            material=melted_alloy_material,
            measurements=[weighting_alloy_measurement],
        )
        block4.link_within()
        block4.link_prior(prior_block, ingredient_name_to_link=alloy_ingredient_name)

        self.subs[block4.name] = block4

        ############## block 5
        homogenization_metadata = processing_details["data"]["Homogenization"][
            "Process Overview"
        ]
        homogenization_performed_source = PerformedSource(
            homogenization_metadata["Completed By"],
            homogenization_metadata["Finish Date"],
        )

        melted_alloy_ingredient_name = "Arc Melted {} Ing.".format(alloy_common_name)
        melted_alloy_ingredient = Ingredient(melted_alloy_ingredient_name)

        homogenization_process = Homogenization(
            "Homogenizing {}".format(alloy_common_name)
        )
        homogenization_tags = (
            ("start_date", homogenization_metadata["Start Date"]),
            ("finish_date", homogenization_metadata["Finish Date"]),
            ("time_spent", str(homogenization_metadata["Time Spent"])),
        )
        homogenization_and_common_tags = common_tags + homogenization_tags
        homogenization_process._set_tags(
            tags=homogenization_and_common_tags,
            spec_or_run=homogenization_process.run,
        )

        def set_homogenization_process_params():
            homogenization_parameters = processing_details["data"]["Homogenization"][
                "Thermal Conditions"
            ]
            homogenization_parameters_2 = processing_details["data"]["Homogenization"][
                "Purging Sequence Pressure"
            ]
            for attribute_name, attribute_value in homogenization_parameters.items():
                if type(attribute_value) == str:
                    value = NominalCategorical(attribute_value)
                else:
                    unit = ""
                    if attribute_name == "Temperature":
                        unit = "Kelvin"
                    elif attribute_name == "Pressure":
                        unit = "Pa"
                    elif attribute_name == "Duration":
                        unit = "hours"
                    value = NominalReal(attribute_value, unit)
                homogenization_process._update_attributes(
                    AttrType=Parameter,
                    attributes=(
                        Parameter(attribute_name, value=value, origin="specified"),
                    ),
                    which="run",
                )
            for purging_step, purging_pressure in homogenization_parameters_2.items():
                if not purging_pressure:
                    continue
                value = NominalReal(purging_pressure, "Pa")
                homogenization_process._update_attributes(
                    AttrType=Parameter,
                    attributes=(
                        Parameter(
                            "Purging Sequence {} Pressure".format(purging_step),
                            value=value,
                            template=homogenization_process._ATTRS["parameters"][
                                "Pressure"
                            ]["obj"],
                            origin="specified",
                        ),
                    ),
                    which="run",
                )
            homogenization_process._update_attributes(
                AttrType=Parameter,
                attributes=(
                    Parameter(
                        "Purging Sequence {} Pressure".format(purging_step),
                        value=value,
                        template=homogenization_process._ATTRS["parameters"][
                            "Pressure"
                        ]["obj"],
                        origin="specified",
                    ),
                ),
                which="run",
            )

        set_homogenization_process_params()
        homogenization_process._run.source = homogenization_performed_source

        homogenized_alloy_material = Alloy("Homogenized {}".format(alloy_common_name))
        homogenized_alloy_material._set_tags(
            tags=common_tags,
            spec_or_run=homogenized_alloy_material.run,
        )
        homogenized_alloy_material._update_attributes(
            AttrType=PropertyAndConditions,
            attributes=(
                PropertyAndConditions(
                    property=Property(
                        "Form",
                        value=NominalCategorical("Ingot"),
                        origin="predicted",
                        notes="",
                        file_links=[],
                    ),
                    conditions=[],
                ),
            ),
            which="spec",
        )

        block5 = ProcessBlock(
            name=f"Homogenization of Alloy {alloy_common_name}",
            workflow=self,
            ingredients=[melted_alloy_ingredient],
            process=homogenization_process,
            material=homogenized_alloy_material,
        )
        block5.link_within()
        block5.link_prior(block4, ingredient_name_to_link=melted_alloy_ingredient_name)
        self.subs[block5.name] = block5

        ############## block 6
        forging_metadata = processing_details["data"]["Forging"]["Process Overview"]
        forging_performed_source = PerformedSource(
            forging_metadata["Completed By"],
            forging_metadata["Finish Date"],
        )
        forging_tags = (
            ("start_date", forging_metadata["Start Date"]),
            ("finish_date", forging_metadata["Finish Date"]),
            ("time_spent", str(forging_metadata["Time Spent"])),
        )

        homogenized_alloy_ingredient_name = "Homogenized {} Ing.".format(
            alloy_common_name
        )
        homogenized_alloy_ingredient = Ingredient(homogenized_alloy_ingredient_name)
        forging_process = Forging("Forging {}".format(alloy_common_name))
        forging_and_common_tags = common_tags + forging_tags
        forging_process._set_tags(
            tags=forging_and_common_tags,
            spec_or_run=forging_process.run,
        )

        def set_forging_process_params():
            attribute_name = "Press Temperature"
            press_temperature = processing_details["data"]["Forging"][attribute_name]
            value = NominalReal(press_temperature, "Kelvin")
            forging_process._update_attributes(
                AttrType=Parameter,
                attributes=(
                    Parameter(
                        attribute_name,
                        value=value,
                        origin="specified",
                        template=forging_process._ATTRS["parameters"]["Temperature"][
                            "obj"
                        ],
                    ),
                ),
                which="run",
            )
            forging_params = processing_details["data"]["Forging"]["Ingot Condition"]
            forging_params_2 = processing_details["data"]["Forging"]["Maximum Load"]
            for attribute_name, attribute_value in forging_params.items():
                unit = ""
                if attribute_name == "Temperature":
                    unit = "Kelvin"
                elif attribute_name == "Soak Time":
                    unit = "minutes"
                value = NominalReal(attribute_value, unit)
                forging_process._update_attributes(
                    AttrType=Parameter,
                    attributes=(
                        Parameter(attribute_name, value=value, origin="specified"),
                    ),
                    which="run",
                )

            for i, maximum_load_step_dict in enumerate(forging_params_2):
                attribute_name = next(iter(maximum_load_step_dict.keys()))
                attribute_value = next(iter(maximum_load_step_dict.values()))

                value = NominalReal(attribute_value, "Pa")
                name = "{} {}".format(attribute_name, i)
                forging_process._update_attributes(
                    AttrType=Parameter,
                    attributes=(
                        Parameter(
                            attribute_name,
                            value=value,
                            origin="specified",
                            template=forging_process._ATTRS["parameters"][
                                "Maximum Load Step"
                            ]["obj"],
                        ),
                    ),
                    which="run",
                )
            forging_process._run.source = forging_performed_source

        set_forging_process_params()

        forged_alloy_material = Alloy("Forged {}".format(alloy_common_name))
        forged_alloy_material._set_tags(
            tags=common_tags,
            spec_or_run=forged_alloy_material.run,
        )

        def set_capture_dimensions_measurement_properties():
            for stage in ["Before", "After"]:
                name = "Ingot Dimensions {}".format(stage)
                _dict = processing_details["data"]["Forging"][name]
                capture_dimensions_measurement = MeasureDimensions(name)
                capture_dimensions_measurement._set_tags(
                    tags=common_tags,
                    spec_or_run=capture_dimensions_measurement.run,
                )
                capture_dimensions_measurement._run.source = forging_performed_source
                for attribute_name, attribute_value in _dict.items():
                    property_name = "{} {}".format(attribute_name, stage)
                    value = NominalReal(attribute_value, "cm")
                    capture_dimensions_measurement._update_attributes(
                        AttrType=Property,
                        attributes=(
                            Property(
                                property_name,
                                value=value,
                                origin="specified",
                                template=capture_dimensions_measurement._ATTRS[
                                    "properties"
                                ][attribute_name]["obj"],
                            ),
                        ),
                        which="run",
                    )

        set_capture_dimensions_measurement_properties()

        block6 = ProcessBlock(
            name=f"Forging of Alloy {alloy_common_name}",
            workflow=self,
            ingredients=[homogenized_alloy_ingredient],
            process=forging_process,
            material=forged_alloy_material,
        )
        block6.link_within()
        block6.link_prior(
            block5, ingredient_name_to_link=homogenized_alloy_ingredient_name
        )
        self.subs[block6.name] = block6

        ############## block 7
        forged_alloy_ingredient_name = "Forged {} Ing.".format(alloy_common_name)
        forged_alloy_ingredient = Ingredient(forged_alloy_ingredient_name)
        setting_traveler_process = SettingTraveler(
            "Setting traveler for {}".format(alloy_common_name)
        )
        setting_traveler_process._set_tags(
            tags=common_tags,
            spec_or_run=setting_traveler_process.run,
        )
        traveler_material = Traveler("{}: Traveler".format(alloy_common_name))
        traveler_material._set_tags(
            tags=common_tags,
            spec_or_run=traveler_material.run,
        )
        block7 = ProcessBlock(
            name=f"Setting up of Traveler {alloy_common_name}",
            workflow=self,
            ingredients=[forged_alloy_ingredient],
            process=setting_traveler_process,
            material=traveler_material,
        )
        block7.link_within()
        block7.link_prior(block6, ingredient_name_to_link=forged_alloy_ingredient_name)
        self.subs[block7.name] = block7
        self.terminal_blocks[composition_id][fabrication_method][batch] = block7

    def process_measurement(
        self,
        item_path,
        composition_id,
        batch,
        fabrication_method,
        bypass=False,
    ):
        """_summary_

        Args:
            item_path (_type_): _description_
            composition_id (_type_): _description_
            batch (_type_): _description_
            fabrication_method (_type_): _description_
            bypass (bool, optional): option to not add the measurement. Defaults to False.
        """
        if bypass:
            return

        not_empty = False
        # check that there is at least one file (!= folder) inside of the item folder
        if os.path.isdir(item_path):
            for p in os.listdir(item_path):
                if not os.path.isdir(os.path.join(item_path, p)):
                    not_empty = True
                    break

        (
            long_name,
            short_name,
            alloy_common_name,
            common_tags,
        ) = return_common_items(composition_id, fabrication_method, batch)

        if not_empty:
            measurement_name = item_path.split("/")[-1]
            measurement_id = item_path.split("/")[-2]
            measurement_obj = self.measurement_types[measurement_name]
            traveler_ingredient_name = "{}: Traveler Ing.".format(alloy_common_name)
            traveler_ingredient = Ingredient(traveler_ingredient_name)
            extract_sample_process = SettingTravelerSample(
                "Extract sample from {}: Traveler".format(alloy_common_name)
            )
            extract_sample_process._set_tags(
                tags=common_tags,
                spec_or_run=extract_sample_process.run,
            )
            traveler_sample = TravelerSample(
                "{}: T. Sample ({}, {})".format(
                    alloy_common_name, measurement_name, measurement_id
                )
            )
            traveler_sample._set_tags(
                tags=common_tags,
                spec_or_run=traveler_sample.run,
            )
            measurement = measurement_obj(
                "{} charact. for {} ({})".format(
                    measurement_name, alloy_common_name, measurement_id
                )
            )
            measurement._set_tags(
                tags=common_tags,
                spec_or_run=measurement.run,
            )

            m = self.measurements[composition_id][fabrication_method][batch][
                measurement_id
            ]
            if m:
                for attribute_name, attribute_value in m.items():
                    attribute_name = attribute_name.replace("/", "_")

                    if not type(attribute_value) == str:  # FIXME
                        unit = measurement._ATTRS["properties"][attribute_name][
                            "obj"
                        ].bounds.default_units
                        value = NominalReal(float(attribute_value), unit)
                    else:
                        value = NominalCategorical(attribute_value)
                    measurement._update_attributes(
                        AttrType=Property,
                        attributes=(
                            Property(
                                attribute_name,
                                value=value,
                                origin="measured",
                                notes="",
                                file_links=[],
                            ),
                        ),
                        which="run",
                    )

            measurement_block = ProcessBlock(
                name="Set {} T. Sample for {} ({}) charact.".format(
                    alloy_common_name, measurement_name, measurement_id
                ),
                workflow=self,
                ingredients=[traveler_ingredient],
                process=extract_sample_process,
                material=traveler_sample,
                measurements=[measurement],
                _type=measurement_name,
            )
            measurement_block.link_within()
            measurement_block.link_prior(
                self.subs[f"Setting up of Traveler {alloy_common_name}"],
                ingredient_name_to_link=traveler_ingredient_name,
            )
            self.subs[measurement_block.name] = measurement_block

            if (
                type(self.terminal_blocks[composition_id][fabrication_method][batch])
                == list
            ):
                self.terminal_blocks[composition_id][fabrication_method][batch].append(
                    measurement_block
                )
            else:
                self.terminal_blocks[composition_id][fabrication_method][batch] = []
                self.terminal_blocks[composition_id][fabrication_method][batch].append(
                    measurement_block
                )

    ###################### HELPERS ###############

    def setup_subfolder(self, composition_id, batch, fabrication_method):
        composition_id_path = os.path.join(self.output / "structured", composition_id)
        if os.path.exists(composition_id_path):
            shutil.rmtree(composition_id_path)
        os.makedirs(composition_id_path)

        fabrication_method_path = os.path.join(composition_id_path, fabrication_method)
        # if not os.path.exists(fabrication_method_path):
        os.makedirs(fabrication_method_path)

        batch_path = os.path.join(fabrication_method_path, batch)
        # if not os.path.exists(batch_path):
        os.makedirs(batch_path)

        raw_jsons_dirpath = os.path.join(batch_path, "raw_jsons")
        thin_jsons_dirpath = os.path.join(batch_path, "thin_jsons")
        os.makedirs(raw_jsons_dirpath)
        os.makedirs(thin_jsons_dirpath)

    def thin_dumps(
        self, obj, destination=None, overwrite=False
    ):  # TODO: add option to pass own target path
        self.local_out_destination = self.output / "terminal_history/thin"
        if destination:  # adding overwrite options
            self.local_out_destination = destination
        if overwrite:
            if os.path.exists(self.local_out_destination):
                shutil.rmtree(self.local_out_destination)
        if not os.path.exists(self.local_out_destination):
            os.makedirs(self.local_out_destination)
        else:  # notifying user that folder is not empty
            if not len(os.listdir(self.local_out_destination)) == 0:
                print("Folder is not empty.")
        print("Executing thin dumps...")
        self.dump_function = self.encoder.thin_dumps
        start = time.time()
        recursive_foreach(obj, self.local_out)
        end = time.time()
        print(f"Time elapsed: {end - start}")

    def dumps(self, obj, destination=None, overwrite=False):
        self.local_out_destination = self.output / "terminal_history/raw"
        if destination:  # adding overwrite options
            self.local_out_destination = destination
        if overwrite:
            if os.path.exists(self.local_out_destination):
                shutil.rmtree(self.local_out_destination)
        if not os.path.exists(self.local_out_destination):
            os.makedirs(self.local_out_destination)
        else:  # notifying user that folder is not empty
            if not len(os.listdir(self.local_out_destination)) == 0:
                print("Folder is not empty.")
        print("Executing raw dumps...")
        self.dump_function = self.encoder.dumps
        start = time.time()
        recursive_foreach(obj, self.local_out)
        end = time.time()
        print(f"Time elapsed: {end - start}")

    def thin_structured_dumps(self):
        """
        dumps the entire model into a JSON per object, each representing the 'thin' version' of the object
        in which pointers (i.e., true value) are replaced by links (e.g., uuid).
        """
        self.dump_function = self.encoder.thin_dumps
        print("Executing thin structured dumps...")
        start = time.time()
        self.structured_dump_loop(mode="thin")
        end = time.time()
        print(f"Time elapsed: {end - start}")

    def structured_dumps(self):
        """
        dumps the entire model into a single JSON, which contains all the model objects with data pointers (!= links).
        """
        self.dump_function = self.encoder.dumps
        print("Executing raw structured dumps...")
        start = time.time()
        self.structured_dump_loop(mode="raw")
        end = time.time()
        print(f"Time elapsed: {end - start}")

    def structured_dump_loop(self, mode="thin"):
        """
        helper function that navigates the blocks of the models and
        """
        for composition_id in gen_compositions(self.root):
            composition_id_path = self.output / "structured_dump" / composition_id
            for fabrication_method in self.terminal_blocks[composition_id].keys():
                if fabrication_method == "DED":
                    continue
                fabrication_method_path = composition_id_path / fabrication_method
                for batch in self.terminal_blocks[composition_id][
                    fabrication_method
                ].keys():
                    self.setup_subfolder(composition_id, batch, fabrication_method)
                    _destination = fabrication_method_path / batch
                    folder_name = "thin_jsons"
                    destination = _destination / folder_name
                    t = self.terminal_blocks[composition_id][fabrication_method][batch]
                    self.local_out_destination = destination  # workaround to: recursive_foreach can't pass params to out()
                    if type(t) == list:
                        for _t in t:
                            if _t.process:
                                for _obj in [_t.process._spec, _t.process._run]:
                                    recursive_foreach(_obj, self.local_out)
                    else:
                        if t.process:
                            for _obj in [t.process._spec, t.process._run]:
                                recursive_foreach(_obj, self.local_out)
                    if self.testing_mode:
                        return

    def link_prior(self, workflow):
        for composition_id in workflow.terminal_blocks.keys():
            for fabrication_method in workflow.terminal_blocks[composition_id].keys():
                for batch in workflow.terminal_blocks[composition_id][
                    fabrication_method
                ].keys():
                    terminal_block = workflow.terminal_blocks[composition_id][
                        fabrication_method
                    ][batch]
                    self.subs[
                        "Infer Compositions"
                    ].material._spec.process = terminal_block.process._spec
                    self.subs[
                        "Infer Compositions"
                    ].material._run.process = terminal_block.process._run
                    # terminal_block.link_posterior(workflow.blocks['Infer Compositions'], ingredient_name_to_link=)

    @classmethod
    def get_command_line_arguments(cls):
        superargs, superkwargs = super().get_command_line_arguments()
        args = [
            *superargs,
            "root",
            "output",
            "iteration",
            "synthesis_path",
            "srjt_path",
        ]
        kwargs = {**superkwargs}
        return args, kwargs

    @classmethod
    def run_from_command_line(cls, args=None):
        parser = cls.get_argument_parser()
        args = parser.parse_args(args=args)
        workflow = cls(
            args.root, args.output, args.iteration, args.synthesis_path, args.srjt_path
        )

        with open(os.path.join(args.output, "log.txt"), "w") as sys.stderr:
            workflow.build()
            workflow.thin_dumps(workflow.terminal_material)
            workflow.thin_structured_dumps()


def gen_compositions(root):
    ids = []
    for id in range(1, 17):
        composition_id = "0%s" % id if id % 10 == id else "%s" % id
        composition_id = str(root).split("/")[-1] + composition_id
        ids.append(composition_id)
    return ids


def return_common_items(composition_id, fabrication_method, batch, yymm=None):
    long_name = (
        f"composition {composition_id} with {fabrication_method} in batch {batch}"
    )
    short_name = f"{composition_id}-{fabrication_method}-{batch}"
    alloy_common_name = "Alloy ({})".format(short_name)
    common_tags = (
        ("composition_id", composition_id),
        ("batch", batch),
        ("fabrication_method", fabrication_method),
    )
    if yymm:
        common_tags = (("yymm", yymm),) + common_tags
    return long_name, short_name, alloy_common_name, common_tags


def ingest_synthesis_results(iteration, synthesis_path, file_links, measurements):
    # file_link_name = "HTMDEC {} Summary Synthesis Results.xlsx".format(iteration)
    # file_link_path = os.path.join(sample_data_folder, file_link_name)
    df = pd.read_excel(synthesis_path)
    file_links["Summary Sheet"] = FileLink(
        filename=str(synthesis_path).split("/")[-1], url=str(synthesis_path)
    )
    new_header = df.iloc[1]
    core_df = df[2:19]
    core_df.columns = new_header
    column_names = list(core_df.columns.values)
    core_df = core_df.reset_index()
    for row_index, row in core_df.iterrows():
        if row_index == 0:
            sub_header_row = row
            continue
        name = row["Alloy"]
        split_name = name.split("_")
        composition_id = split_name[0]
        yymm = split_name[1]
        fabrication_method = split_name[2]
        batch = split_name[3]

        # target composition
        target_composition_column = column_names[1]
        for i in range(2, 8):
            element_name = sub_header_row[i]
            composition = row[i]
            measurements[composition_id][target_composition_column][
                element_name
            ] = composition

        # T05: averaged measured composition and difference
        measurement_id = row[8]
        average_composition_column = column_names[8]
        composition_difference_column = column_names[14]
        for i in range(9, 15):
            element_name = sub_header_row[i]
            average_composition = row[i]
            composition_difference = row[i + 6]
            measurements[composition_id][fabrication_method][batch][measurement_id][
                average_composition_column
            ][element_name] = average_composition
            measurements[composition_id][fabrication_method][batch][measurement_id][
                composition_difference_column
            ][element_name] = composition_difference

        # T03: phase/lattice parameters
        measurement_id = row[21]
        phase_column = column_names[21]
        lattice_parameter_column = column_names[22]
        phase = row[22]
        lattice_parameter = row[23]

        measurements[composition_id][fabrication_method][batch][measurement_id][
            phase_column
        ] = phase
        measurements[composition_id][fabrication_method][batch][measurement_id][
            lattice_parameter_column
        ] = lattice_parameter

        # T03: hardness, HV, SD, HV
        measurement_id = row[24]
        hardness_column = column_names[24]
        sd_hv_column = column_names[25]
        hardness = row[25]
        sd_hv = row[26]
        measurements[composition_id][fabrication_method][batch][measurement_id][
            hardness_column
        ] = hardness
        measurements[composition_id][fabrication_method][batch][measurement_id][
            sd_hv_column
        ] = sd_hv

        # T08: elastic modulus, yield strength, uts, elongtation, strain hardnening
        measurement_id = row[27]
        elastic_modulus_column = column_names[27]
        yield_stength_column = column_names[28]
        uts_column = column_names[29]
        elongation_column = column_names[30]
        strain_hardening_column = column_names[31]
        derivative_column = column_names[32]
        derivative_column = derivative_column.replace("\u03c3", "d")

        elastic_modulus = row[28]
        yield_stength = row[29]
        uts = row[30]
        elongation = row[31]
        strain_hardening = row[32]
        derivative = row[33]

        measurements[composition_id][fabrication_method][batch][measurement_id][
            elastic_modulus_column
        ] = elastic_modulus
        measurements[composition_id][fabrication_method][batch][measurement_id][
            yield_stength_column
        ] = yield_stength
        measurements[composition_id][fabrication_method][batch][measurement_id][
            uts_column
        ] = uts
        measurements[composition_id][fabrication_method][batch][measurement_id][
            elongation_column
        ] = elongation
        measurements[composition_id][fabrication_method][batch][measurement_id][
            strain_hardening_column
        ] = strain_hardening
        measurements[composition_id][fabrication_method][batch][measurement_id][
            derivative_column
        ] = derivative

        # T09: elastic modulus, yield strength, uts, elongtation, strain hardnening
        measurement_id = row[34]
        elastic_modulus_column = column_names[34]
        yield_stength_column = column_names[35]
        uts_column = column_names[36]
        elongation_column = column_names[37]
        strain_hardening_column = column_names[38]
        derivative_column = column_names[39]
        derivative_column = derivative_column.replace("\u03c3", "d")

        elastic_modulus = row[35]
        yield_stength = row[36]
        uts = row[37]
        elongation = row[38]
        strain_hardening = row[39]
        derivative = row[40]

        measurements[composition_id][fabrication_method][batch][measurement_id][
            elastic_modulus_column
        ] = elastic_modulus
        measurements[composition_id][fabrication_method][batch][measurement_id][
            yield_stength_column
        ] = yield_stength
        measurements[composition_id][fabrication_method][batch][measurement_id][
            uts_column
        ] = uts
        measurements[composition_id][fabrication_method][batch][measurement_id][
            elongation_column
        ] = elongation
        measurements[composition_id][fabrication_method][batch][measurement_id][
            strain_hardening_column
        ] = strain_hardening
        measurements[composition_id][fabrication_method][batch][measurement_id][
            derivative_column
        ] = derivative

        # Average: elastic modulus, yield strength, uts, elongtation, strain hardnening
        measurement_id = row[41]
        elastic_modulus_column = column_names[41]
        yield_stength_column = column_names[42]
        uts_column = column_names[43]
        elongation_column = column_names[44]
        strain_hardening_column = column_names[45]
        derivative_column = column_names[46]
        derivative_column = derivative_column.replace("\u03c3", "d")

        elastic_modulus = row[42]
        yield_stength = row[43]
        uts = row[44]
        elongation = row[45]
        strain_hardening = row[46]
        derivative = row[47]

        measurements[composition_id][fabrication_method][batch][measurement_id][
            elastic_modulus_column
        ] = elastic_modulus
        measurements[composition_id][fabrication_method][batch][measurement_id][
            yield_stength_column
        ] = yield_stength
        measurements[composition_id][fabrication_method][batch][measurement_id][
            uts_column
        ] = uts
        measurements[composition_id][fabrication_method][batch][measurement_id][
            elongation_column
        ] = elongation
        measurements[composition_id][fabrication_method][batch][measurement_id][
            strain_hardening_column
        ] = strain_hardening
        measurements[composition_id][fabrication_method][batch][measurement_id][
            derivative_column
        ] = derivative

    strain_rate_and_temperature_df = df[19:21]
    strain_rate_and_temperature_df.columns = new_header


def ingest_srjt_results(srjt_path, output, measurements):
    # FIXME: add file link
    # Convert the Excel file to CSV using LibreOffice
    output_file = str(srjt_path).split(".")[0] + ".csv"
    output_file = output_file.split("/")[-1]

    conversion_command = [
        "libreoffice",
        "--headless",  # Run LibreOffice in headless mode (without GUI)
        "--convert-to",
        "csv",
        "--outdir",
        output,  # Output directory
        srjt_path,
    ]

    subprocess.run(conversion_command)

    # Read the CSV file into a Pandas DataFrame
    df = pd.read_csv(output / output_file)

    # Clean up the temporary CSV file
    subprocess.run(["rm", output / output_file])

    columns = list(df.columns)
    # sample_name_coulmn =
    for i in range(len(df)):
        composition_id = df.loc[i, "Sample Name"]
        for y in range(1, len(columns)):
            measurements[composition_id]["SRJT"][columns[y]] = df.loc[i, columns[y]]


def substring_after(s, delim):
    return s.partition(delim)[2]


def main(args=None):
    """
    Main method to run from command line
    """
    BIRDSHOTWorfklow.run_from_command_line(args)


if __name__ == "__main__":
    main()



from typing import ClassVar

from gemd import (
    ProcessTemplate,
    ParameterTemplate,
    CategoricalBounds,
    NominalCategorical,
    RealBounds,
    NominalReal,
)

from openmsimodel.entity.base.process import Process

from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["Homogenization"]


class Homogenization(Process):
    """Class representing homogenization cycle"""

    TEMPLATE: ClassVar[ProcessTemplate] = ProcessTemplate(
        name="Homogenization",
        description="""Homogenizing
                """,
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}}

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Pressure",
            bounds=RealBounds(0, 20, "Pa"),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Duration",
            bounds=RealBounds(0, 72, "hours"),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Temperature",
            bounds=RealBounds(0, 2000, "Kelvin"),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Atmosphere",
            bounds=CategoricalBounds(["Ar"]),
        ),
    )

    define_attribute(
        _ATTRS,
        template=ParameterTemplate(
            name="Cooling Rate",
            bounds=CategoricalBounds(["FC"]),
        ),
    )

    finalize_template(_ATTRS, TEMPLATE)



from typing import Callable, Union, Type, Tuple
from gemd.entity.link_by_uid import LinkByUID
from gemd.entity.dict_serializable import DictSerializable

from gemd.entity.object import MaterialRun, ProcessRun, IngredientRun, MeasurementRun

# from openmsimodel.entity.base.base_element import BaseElement
from openmsimodel.utilities.typing import (
    Spec,
    Run,
    Template,
    AttrTemplate,
    ObjTemplate,
    AllGEMD,
    AllGEMDMinusAttr,
)


######### THE FUNCTION BELOW IS COPIED/PASTED FROM GEMD CODE #########
def cached_isinstance_generator(
    class_or_tuple: Union[Type, Tuple[Type]]
) -> Callable[[object], bool]:
    """
    Generate a function that checks and caches an isinstance(obj, class_or_tuple) call.

    Parameters
    ----------
    class_or_tuple: Union[Type, Tuple[Type]]
        A single type or a tuple of types

    Returns
    -------
    Callable[[object], bool]
        function with signature function(obj), returning isinstance(obj, class_or_tuple)

    """
    cache = dict()

    def func(obj):
        obj_type = type(obj)
        if obj_type not in cache:
            cache[obj_type] = isinstance(obj, class_or_tuple)
        return cache[obj_type]

    return func


# from openmsimodel.entity.base.process import Process
# from openmsimodel.entity.base.material import Material
# from openmsimodel.utilities.attributes import _validate_temp_keys, define_attribute

# Some cached isinstance functions to reduce overhead
# isinstance_base_element = cached_isinstance_generator([BaseElement])

isinstance_all_gemd = cached_isinstance_generator(AllGEMD.__args__)
isinstance_all_gemd_minus_attr = cached_isinstance_generator(AllGEMDMinusAttr.__args__)

isinstance_template = cached_isinstance_generator(Template.__args__)
isinstance_attribute_template = cached_isinstance_generator(AttrTemplate.__args__)
isinstance_object_template = cached_isinstance_generator(ObjTemplate.__args__)
isinstance_spec = cached_isinstance_generator(Spec.__args__)
isinstance_run = cached_isinstance_generator(Run.__args__)

isinstance_link_by_uid = cached_isinstance_generator(LinkByUID)
isinstance_dict_serializable = cached_isinstance_generator(DictSerializable)

isinstance_material_run = cached_isinstance_generator(MaterialRun)
isinstance_ingredient_run = cached_isinstance_generator(IngredientRun)

isinstance_list_or_tuple = cached_isinstance_generator((list, tuple))



from typing import ClassVar

from gemd import MeasurementTemplate

from openmsimodel.entity.base.measurement import Measurement
from openmsimodel.utilities.attributes import (
    AttrsDict,
    define_attribute,
    finalize_template,
)

__all__ = ["NI"]


class NI(Measurement):
    """Class representing NI as a measurement"""

    TEMPLATE: ClassVar[MeasurementTemplate] = MeasurementTemplate(
        name="NI", description="NI measurement"
    )

    _ATTRS: ClassVar[AttrsDict] = {"conditions": {}, "parameters": {}, "properties": {}}

    # define_attribute(
    #     _ATTRS,
    #     template=ParameterTemplate(
    #         name='Supplier',
    #         bounds=CategoricalBounds(categories=[''])
    #     ),
    #     default_value=NominalCategorical('')
    # )

    finalize_template(_ATTRS, TEMPLATE)
